import dataclasses
import datetime
import pandas as pd
import numpy as np
import math
import statsmodels.api as sm
from typing import Optional
from statsmodels import regression
from scipy.stats import gmean
from functools import partial
import statsmodels.api as sm
from statsmodels.tsa.ar_model import AutoReg
from scipy.stats import kurtosis, skew
from decimal import Decimal, ROUND_HALF_UP, Context
from pandas.tseries.offsets import DateOffset
from ..data.api.basic import BasicDataApi

class CalculatorBase:

    TRADING_DAYS_PER_YEAR = 242
    TOTAL_DAYS_PER_YEAR = 365
    TOTAL_WEEKS_PER_YEAR = 52
    TOTAL_MONTHS_PER_YEAR = 12
    STOCK_IPO_NUM_MARKET = {
        '深交所中小板': 500,
        '深交所创业板': 500,
        '上交所主板': 1000,
        '上交所科创板': 500,
        '深交所主板': 500,
    }

    @staticmethod
    def get_begin_date_dic(week_end_date):
        begin_date_period_dic = {
            'w1': DateOffset(weeks=1),
            'm1': DateOffset(months=1),
            'm3': DateOffset(months=3),
            'm6': DateOffset(months=6),
            'y1': DateOffset(years=1),
            'y2': DateOffset(years=2),
            'y3': DateOffset(years=3),
            'y4': DateOffset(years=4),
            'y5': DateOffset(years=5),
        }
        last_date = datetime.datetime.now().date()
        begin_date_dic = {}
        for name, value in begin_date_period_dic.items():
            if name == 'w1':
                bd = (week_end_date - value).date()
            else:
                bd = (week_end_date - value).date()
            begin_date_dic[name] = bd
        return last_date, begin_date_dic

    @staticmethod
    def calc_cl_alpha_beta(total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 0][X[:, 0] < 0] = 0
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'beta_bull':np.nan,
                    'beta_bear':np.nan,
                    
                    }
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                'beta_bull':est2.params[0],
                'beta_bear':-est2.params[1],
                }
    
    @staticmethod
    def calc_hm_alpha_beta(total: np.ndarray):
        # 牛市追踪 市场上涨是alpha
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 1][X[:, 1] < 0] = 0
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[1],
                'alpha':est2.params[-1],
                }

    @staticmethod
    def stat_round_pct(x):
        s = str(round(100*x,2))
        s = s[::-1].zfill(4)[::-1]
        return s + '%'

    @staticmethod
    def calc_continue_regress_v(monthly_ret, risk_free_rate):
        # 半年周期
        window = 6
        yearly_multiplier = 2
        period_num = 12
        annual_ret = monthly_ret.rolling(window=window).apply(lambda x : x.sum() * yearly_multiplier - risk_free_rate)
        annual_vol = monthly_ret.rolling(window=window).apply(lambda x : x.std(ddof=1) * np.sqrt(period_num))
        sharpe = (annual_ret / annual_vol)
        dic = {np.Inf:np.nan,-np.Inf:np.nan}
        sharpe = sharpe.replace(dic).dropna()
        if sharpe.shape[0] < 6:
            return np.nan
        mod = AutoReg(endog=sharpe.values, lags=1)
        res = mod.fit()
        continue_regress_v = res.params[0]
        return continue_regress_v

    @staticmethod
    def calc_modify_annual_vol(sr: pd.core.series.Series):
        # 每个ret(i)除以产生它的时长平方根sqrt(T(i))后，服从独立同分布，可以用来估计日波动率，最后再乘以sqrt(242)年化
        dts = BasicDataApi().get_trading_day_list(start_date=sr.index[0], end_date=sr.index[-1]).set_index('datetime')
        sr_ret = sr.pct_change(1)
        sr_ret.name = 'ret'
        sr_ret.index.name = 'this_day'
        sr_ret = sr_ret.reset_index()
        sr_ret.loc[:,'last_day'] = sr_ret.this_day.shift(1)
        sr_ret = sr_ret.dropna()
        sr_ret.loc[:,'trade_dates'] = sr_ret.apply(lambda x: dts.loc[x.last_day:x.this_day].shape[0], axis=1)
        sr_ret.loc[:,'modify_ret'] = sr_ret.ret / np.sqrt(sr_ret.trade_dates)
        annual_vol = sr_ret.modify_ret.replace({np.inf:np.nan,-np.inf:np.nan}).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
        return annual_vol

    @staticmethod
    def calc_recent_ret(dates:             pd.core.series.Series,
                        values:            pd.core.series.Series,
                        frequency:         str='1D'):
        try:
            assert frequency in ['1D'], f'frequency provided: {frequency} is not included, must in 1D'
            dates = pd.to_datetime(dates)
            dates = [i.date() for i in dates]
            _end_dt = dates[-1]
            last_date, begin_date_dic = CalculatorBase.get_begin_date_dic(week_end_date = _end_dt)
            assert len(dates) == len(values), 'date and port_values has different numbers'
            sr = pd.Series(values, index=dates).sort_index().dropna()
            sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency).last().dropna()
            sr.index = [i.date() for i in sr.index]
            _begin_date = sr.index[0]
            _natural_year_begin_date = datetime.date(last_date.year,1,1)
            _sr = sr.loc[:_natural_year_begin_date]
            if _sr.shape[0] > 2:
                _natural_year_begin_date = _sr.index[-2] if _natural_year_begin_date in sr.index else _sr.index[-1]
                recent_natural_year_ret = sr.loc[:last_date][-1] / sr.loc[_natural_year_begin_date] - 1
            else:
                _sr = sr.loc[_natural_year_begin_date:]
                if _sr.empty:
                    recent_natural_year_ret = 0
                else:
                    _natural_year_begin_date = _sr.index[0]
                    recent_natural_year_ret = sr.loc[:last_date][-1] / sr.loc[_natural_year_begin_date] - 1
            _res = {}
            for _period, _period_begin_date in begin_date_dic.items():
                if _period_begin_date >= _begin_date and sr.index[-1] > _period_begin_date:
                    _res[_period] = sr.loc[:last_date][-1] / sr.loc[_period_begin_date:][0] - 1
                    if _period == 'w1':
                        _res['w1_begin_date'] = sr.loc[_period_begin_date:].index[0]
                        _res['w1_end_date'] = sr.loc[:last_date].index[-1]
                else:
                    _res[_period] = np.nan
                    
            recent_1w_ret = _res['w1']
            recent_1m_ret = _res['m1']
            recent_3m_ret = _res['m3']
            recent_6m_ret = _res['m6']
            recent_1y_ret = _res['y1']
            recent_2y_ret = _res['y2']
            recent_3y_ret = _res['y3']
            recent_4y_ret = _res['y4']
            recent_5y_ret = _res['y5']
            if 'w1_begin_date' in _res:
                w1_begin_date = _res['w1_begin_date']
                w1_end_date = _res['w1_end_date']
            else:
                w1_begin_date = sr.index[0]
                w1_end_date = sr.index[-1]

            history_ret = sr.loc[:last_date][-1] / sr[0] - 1
            days = (dates[-1] - dates[0]).days
            history_annual_ret = math.pow(history_ret + 1, CalculatorBase.TOTAL_DAYS_PER_YEAR/days) - 1 if days != 0 else 0
            sr_week = CalculatorBase.data_resample_weekly_nav(sr,rule='W-FRI')
            if sr_week.shape[0] < 5:
                recent_4w_ret = sr_week[-1] / sr_week[0] - 1
            else:
                recent_4w_ret = sr_week[-1] / sr_week[-5] - 1

            _date_20_b = datetime.date(2019,12,31)
            _date_20_e = datetime.date(2020,12,31)
            date_20_b = max(sr.index.min(),_date_20_b)
            date_20_e = min(sr.index.max(),_date_20_e)
            _sr = sr[date_20_b:date_20_e]
            y_2020_ret = _sr[-1] / _sr[0] - 1 if not _sr.empty else None
            lasty_year = sr.index[-1].year - 1
            _date_last_b = datetime.date(lasty_year-1,12,31)
            _date_last_e = datetime.date(lasty_year,12,31)
            date_last_b = max(sr.index.min(),_date_last_b)
            date_last_e = min(sr.index.max(),_date_last_e)
            _sr = sr[date_last_b:date_last_e]
            y_last_ret = _sr[-1] / _sr[0] - 1 if not _sr.empty else None
            return {
                'recent_natural_year_ret':recent_natural_year_ret,
                'recent_1w_ret':recent_1w_ret,
                'recent_4w_ret':recent_4w_ret,
                'recent_1m_ret':recent_1m_ret,
                'recent_3m_ret':recent_3m_ret,
                'recent_6m_ret':recent_6m_ret,
                'recent_1y_ret':recent_1y_ret,
                'recent_2y_ret':recent_2y_ret,
                'recent_3y_ret':recent_3y_ret,
                'recent_4y_ret':recent_4y_ret,
                'recent_5y_ret':recent_5y_ret,
                'history_ret':history_ret,
                'history_annual_ret':history_annual_ret,
                'y_2020_ret':y_2020_ret,
                'last_year_ret':y_last_ret,
                'w1_begin_date':w1_begin_date,
                'w1_end_date':w1_end_date,
            }
        
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.calc_recent_ret')
            return {
                'recent_natural_year_ret':0,
                'recent_1w_ret':0,
                'recent_4w_ret':0,
                'recent_1m_ret':0,
                'recent_3m_ret':0,
                'recent_6m_ret':0,
                'recent_1y_ret':0,
                'recent_2y_ret':0,
                'recent_3y_ret':0,
                'recent_4y_ret':0,
                'recent_5y_ret':0,
                'history_ret':0,
                'history_annual_ret':0,
                'y_2020_ret':0,
                'last_year_ret':0,
                'w1_begin_date':None,
                'w1_end_date':None,
            }

    @staticmethod
    def get_period_ret(dates: pd.core.series.Series,
                       values: pd.core.series.Series,
                       frequency: str='3M'):
        assert frequency in ['1D','1W','2W','1M','3M','1Y'], 'frequency provided is not included, must in 1D 1W 2W 1M 3M 1Y'
        if frequency == '1Y':
            sr = pd.Series(values, index=dates).sort_index().dropna()
            year_list = set(sorted([i.year for i in sr.index]))
            _date = []
            for y in year_list:
                for m_d in [[12,31]]:
                    m = m_d[0]
                    d = m_d[1]
                    dt = datetime.date(y,m,d)
                    _date.append(dt)
            sr = sr.reindex(sr.index.union(_date)).sort_index().ffill().bfill()
            sr = sr.reindex([sr.index[0]] + _date).pct_change(1).dropna()
            sr = sr[sr!=0]
            df = pd.DataFrame(sr)
            df.columns = ['ret']
            return df.sort_index()
        if frequency == '3M':
            sr = pd.Series(values, index=dates).sort_index().dropna()
            year_list = set(sorted([i.year for i in sr.index]))
            _date = []
            for y in year_list:
                for m_d in [[3,31],[6,30],[9,30],[12,31]]:
                    m = m_d[0]
                    d = m_d[1]
                    dt = datetime.date(y,m,d)
                    _date.append(dt)
            sr = sr.reindex(sr.index.union(_date)).sort_index().ffill().bfill()
            sr = sr.reindex(_date).pct_change(1).dropna()
            sr = sr[sr!=0]
            df = pd.DataFrame(sr)
            df.columns = ['ret']
            return df.sort_index()
        sr = pd.Series(values, index=dates).sort_index().dropna()
        sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency,).last().dropna()
        df = pd.DataFrame(sr.pct_change(1).dropna())
        df.columns = ['ret']
        td = df.index
        df.index = [i.date() for i in td]
        return df.sort_index()

    @staticmethod
    def calc_recent_vol(dates:             pd.core.series.Series,
                        values:            pd.core.series.Series,
                        frequency:         str='1D'):
        assert frequency in ['1D','1W'], 'frequency provided is not included, must in 1D 1W 2W 1M'
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        _end_dt = dates[-1]
        last_date, begin_date_dic = CalculatorBase.get_begin_date_dic(week_end_date = _end_dt)
        assert len(dates) == len(values), 'date and port_values has different numbers'
        sr = pd.Series(values, index=dates).sort_index().dropna()
        sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency).last().dropna()
        sr.index = [i.date() for i in sr.index]
        sr_ret = sr.pct_change(1).dropna()
        if frequency == '1D':    
            period_1y = CalculatorBase.TRADING_DAYS_PER_YEAR
        elif frequency == '1W':
            period_1y = CalculatorBase.TOTAL_WEEKS_PER_YEAR
        _begin_date = sr_ret.index[0]
        _natural_year_begin_date = datetime.date(last_date.year,1,1)
        recent_natural_year_vol = sr_ret.loc[_natural_year_begin_date:].std(ddof=1) * np.sqrt(period_1y)
        recent_1w_vol = sr_ret.loc[max(begin_date_dic['w1'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_1m_vol = sr_ret.loc[max(begin_date_dic['m1'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_3m_vol = sr_ret.loc[max(begin_date_dic['m3'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_6m_vol = sr_ret.loc[max(begin_date_dic['m6'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_1y_vol = sr_ret.loc[max(begin_date_dic['y1'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_3y_vol = sr_ret.loc[max(begin_date_dic['y3'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_5y_vol = sr_ret.loc[max(begin_date_dic['y5'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        history_vol = sr_ret.std(ddof=1) * np.sqrt(period_1y)
        return {
            'recent_natural_year_vol':recent_natural_year_vol,
            'recent_1w_vol':recent_1w_vol,
            'recent_1m_vol':recent_1m_vol,
            'recent_3m_vol':recent_3m_vol,
            'recent_6m_vol':recent_6m_vol,
            'recent_1y_vol':recent_1y_vol,
            'recent_3y_vol':recent_3y_vol,
            'recent_5y_vol':recent_5y_vol,
            'history_vol':history_vol,
        }

    @staticmethod
    def get_period_vol(dates: pd.core.series.Series,
                       values: pd.core.series.Series,
                       frequency: str='3M'):
        assert frequency in ['1D','1W','2W','1M','3M','1Y'], 'frequency provided is not included, must in 1D 1W 2W 1M 3M 1Y'
        sr = pd.Series(values, index=dates).sort_index().pct_change(1).dropna()
        if frequency == '1D':
            roll_d = CalculatorBase.TRADING_DAYS_PER_YEAR
        elif frequency == '1W':
            roll_d = 5
        elif frequency == '2W':
            roll_d = 10
        elif frequency == '1M':
            roll_d = 20
        elif frequency == '3M':
            roll_d = 60
        elif frequency == '1Y':
            roll_d = CalculatorBase.TRADING_DAYS_PER_YEAR
        if frequency == '1Y':
            sr = pd.Series(values, index=dates).sort_index().dropna()
            year_list = set(sorted([i.year for i in sr.index]))
            _date = []
            for y in year_list:
                for m_d in [[12,31]]:
                    m = m_d[0]
                    d = m_d[1]
                    dt = datetime.date(y,m,d)
                    _date.append(dt)
            sr = sr.reindex(sr.index.union(_date)).sort_index().ffill().pct_change(1)
            sr = sr.rolling(window=roll_d).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
            sr = sr.reindex(_date).dropna()
            sr = sr[sr!=0]
            df = pd.DataFrame(sr)
            df.columns = ['vol']
            return df.sort_index()
        if frequency == '3M':
            sr = pd.Series(values, index=dates).sort_index().dropna()
            year_list = set(sorted([i.year for i in sr.index]))
            _date = []
            for y in year_list:
                for m_d in [[3,31],[6,30],[9,30],[12,31]]:
                    m = m_d[0]
                    d = m_d[1]
                    dt = datetime.date(y,m,d)
                    _date.append(dt)
            sr = sr.reindex(sr.index.union(_date)).sort_index().ffill().pct_change(1)
            sr = sr.rolling(window=roll_d).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
            sr = sr.reindex(_date).dropna()
            sr = sr[sr!=0]
            df = pd.DataFrame(sr)
            df.columns = ['vol']
            df = df.sort_index()
            today = datetime.datetime.now().date()
            this_year = today.year
            year_list = np.array([datetime.date(this_year, m_d[0], m_d[1]) for m_d in [[3,31],[6,30],[9,30],[12,31]]])
            next_report_date = year_list[year_list>today][0]
            return df.loc[:next_report_date]
        sr = sr.rolling(window=roll_d).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
        sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency).last().dropna()
        df = pd.DataFrame(sr)
        df.columns = ['vol']
        td = df.index
        df.index = [i.date() for i in td]
        return df.sort_index()

    @staticmethod
    def get_fund_correlation_df(df: pd.DataFrame):
        # 列是资产代码 index是日期
        # 求最近一年日度收益相关性
        return df.pct_change(1).corr()

    @staticmethod
    def get_stat_result(dates:             pd.core.series.Series,
                        values:            pd.core.series.Series,
                        benchmark_values:  pd.core.series.Series=None,
                        risk_free_rate:    float=0.025,
                        frequency:         str='1D',
                        ret_method:        str='log_ret',
                        ):# '1W' '1M'
        assert frequency in ['1D','1W','2W','1M'], 'frequency provided is not included, must in 1D 1W 2W 1M'
        assert ret_method in ['log_ret', 'pct_ret'], 'ret_method is not included'
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        _end_dt = dates[-1]
        last_date, begin_date_dic = CalculatorBase.get_begin_date_dic(week_end_date = _end_dt)
        assert len(dates) == len(values), 'date and port_values has different numbers'
        if len(dates) >= 2:
            sr = pd.Series(values, index=dates).sort_index()
            sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency).last().dropna()
            sr.index = [i.date() for i in sr.index]
            natural_day_lenth = (sr.index[-1] - sr.index[0]).days
            price_num = sr.dropna().shape[0]
            nav_density = price_num / natural_day_lenth
            nav_type = CalculatorBase.get_nav_type(values=values, dates=dates)
            sr = sr.dropna()
            if frequency == '1D':            
                period_1m = 20
                period_3m = period_1m * 3
                period_6m = period_1m * 6
                period_1y = CalculatorBase.TRADING_DAYS_PER_YEAR
                risk_free_per_period = risk_free_rate / CalculatorBase.TRADING_DAYS_PER_YEAR
            elif frequency == '1W':
                period_1m = 4
                period_3m = period_1m * 3
                period_6m = period_1m * 6
                period_1y = CalculatorBase.TOTAL_WEEKS_PER_YEAR
                risk_free_per_period = risk_free_rate / 52
            elif frequency == '2W':
                period_1m = 2
                period_3m = period_1m * 3
                period_6m = period_1m * 6
                period_1y = CalculatorBase.TOTAL_WEEKS_PER_YEAR / 2
                risk_free_per_period = risk_free_rate / 26
            elif frequency == '1M':
                period_1m = 1
                period_3m = period_1m * 3
                period_6m = period_1m * 6
                period_1y = CalculatorBase.TOTAL_MONTHS_PER_YEAR
                risk_free_per_period = risk_free_rate / 12
            if ret_method == 'log_ret':
                sr_ret = np.log(sr).diff(1).iloc[1:]
            elif ret_method == 'pct_ret':
                sr_ret = sr.pct_change(1).iloc[1:]
            start_date = sr.index[0] 
            end_date = sr.index[-1]
            days = (dates[-1] - dates[0]).days
            last_unit_nav = sr[-1] / sr[0]
            cumu_ret = last_unit_nav - 1
            trade_year = days / CalculatorBase.TOTAL_DAYS_PER_YEAR
            annual_ret = math.pow(last_unit_nav, CalculatorBase.TOTAL_DAYS_PER_YEAR/days) - 1
            annual_vol = CalculatorBase.calc_modify_annual_vol(sr)
            _begin_date = sr.index[0]
            recent_ret_res = CalculatorBase.calc_recent_ret(dates=dates,values=values,frequency='1D')
            recent_year_ret = recent_ret_res['recent_natural_year_ret']
            recent_1w_ret = recent_ret_res['recent_1w_ret']
            recent_1m_ret = recent_ret_res['recent_1m_ret']
            recent_3m_ret = recent_ret_res['recent_3m_ret']
            recent_6m_ret = recent_ret_res['recent_6m_ret']
            recent_1y_ret = recent_ret_res['recent_1y_ret']
            recent_3y_ret = recent_ret_res['recent_3y_ret']
            recent_5y_ret = recent_ret_res['recent_5y_ret']
            recent_4w_ret = recent_ret_res['recent_4w_ret']
            last_year_ret = recent_ret_res['last_year_ret']
            w1_begin_date = recent_ret_res['w1_begin_date']
            w1_end_date = recent_ret_res['w1_end_date']
            y_2020_ret = recent_ret_res['y_2020_ret']
            worst_3m_ret = sr.pct_change(period_3m).min()
            worst_6m_ret = sr.pct_change(period_6m).min()
            last_mv_diff = sr[-1] - sr[-2]
            last_increase_rate = (sr[-1] - sr[-2])/ sr[-2]
            sharpe = (annual_ret - risk_free_rate) / annual_vol
            recent_drawdown = 1 - (sr[-1] / sr.max())
            if np.isnan(recent_drawdown):
                recent_mdd_date1 = None
                recent_mdd_lens = None
            else:
                recent_mdd_date1 = sr.idxmax()
                recent_mdd_lens = (end_date - recent_mdd_date1).days
            mdd_part =  sr / sr.rolling(window=sr.shape[0], min_periods=1).max()
            mdd = 1 - mdd_part.min()
            if np.isnan(mdd):
                mdd_date2 = None
                mdd_date1 = None
                mdd_lens = None
            else:
                mdd_date2 = mdd_part.idxmin()
                mdd_date1 = sr[:mdd_date2].idxmax()
                mdd_lens = (mdd_date2-mdd_date1).days
            calmar = (annual_ret - risk_free_rate) / mdd if mdd != 0 else 0
            sr_ret_no_risk_free = sr_ret - risk_free_rate / period_1y
            downside_risk = np.minimum(sr_ret_no_risk_free,0).std(ddof=1) * np.sqrt(period_1y)
            sortino = (annual_ret - risk_free_rate) / downside_risk
            var =  np.quantile(sr_ret, 0.05) # 按照最大回撤处理， 损失非负， 计算ervar用
            cvar = max(0, -sr_ret[sr_ret < var].mean())
            var = - var
            ervar = (annual_ret - risk_free_rate) / var
            skew_value = skew(sr_ret)
            kurtosis_value = kurtosis(sr_ret)        
            sr_ret_monthly = CalculatorBase.data_resample_monthly_ret(df=sr_ret, min_count=1).dropna()
            raise_month_num = sr_ret_monthly[sr_ret_monthly>0].shape[0]
            drop_month_num = sr_ret_monthly[sr_ret_monthly<0].shape[0]
            mdd_recover, mdd_recover_date1, mdd_recover_date2, mdd_recover_lens = CalculatorBase.mdd_recover_analysis(values=values, dates=dates)
            win_rate_0_period = np.mean((sr_ret > 0) * 1)

            if benchmark_values is not None:
                assert len(dates) == len(benchmark_values), 'date and bench_values has different numbers'
                sr_benchmark = pd.Series(benchmark_values, index=dates).sort_index().dropna()
                sr_benchmark = sr_benchmark.set_axis(pd.to_datetime(sr_benchmark.index), inplace=False).resample(frequency).last().dropna()
                total_relative_ret = sr.iloc[-1]/ sr.iloc[0] - sr_benchmark.iloc[-1] / sr_benchmark.iloc[0]
                sr_benchmark.index = [i.date() for i in sr_benchmark.index]
                if ret_method == 'log_ret':
                    sr_benchmark_ret = np.log(sr_benchmark).diff(1).iloc[1:]
                elif ret_method == 'pct_ret':
                    sr_benchmark_ret = sr_benchmark.pct_change(1).iloc[1:]
                last_unit_benchmark = sr_benchmark[-1] / sr_benchmark[0]
                annual_ret_benchmark = math.pow(last_unit_benchmark, CalculatorBase.TOTAL_DAYS_PER_YEAR / days) - 1
                df = pd.concat([sr, sr_benchmark],axis=1)
                df.columns= ['fund_ret','index_ret']
                df_monthly = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule='1M').last()
                df_monthly.index = df_monthly.index.date
                monthly_ret = df_monthly.pct_change(1).dropna()
                continue_value_m = CalculatorBase.calc_continue_regress_v(monthly_ret.fund_ret,risk_free_rate)
                df_ret_risk_free = (df.pct_change(1) - risk_free_per_period).dropna()
                cl_res = CalculatorBase.calc_cl_alpha_beta(df_ret_risk_free.to_numpy())
                alpha_cl = cl_res['alpha'] * period_1y
                beta_cl = cl_res['beta']
                beta_bear = cl_res['beta_bear']
                beta_bull = cl_res['beta_bull']
                hm_res = CalculatorBase.calc_hm_alpha_beta(df_ret_risk_free.to_numpy())
                alpha_hm = hm_res['alpha'] * period_1y
                beta_hm = hm_res['beta']
                excess_ret = annual_ret - annual_ret_benchmark
                _index = sorted(sr_ret.index.intersection(sr_benchmark_ret.index))
                sr_ret = sr_ret.reindex(_index)
                sr_benchmark_ret = sr_benchmark_ret.reindex(_index)
                sr_ret_weekly = CalculatorBase.data_resample_weekly_ret(df=sr_ret,min_count=1).dropna()
                sr_benchmark_ret_weekly = CalculatorBase.data_resample_weekly_ret(df=sr_benchmark_ret, min_count=1).dropna()
                sr_benchmark_ret_monthly = CalculatorBase.data_resample_monthly_ret(df=sr_benchmark_ret, min_count=1).dropna()
                
                
                win_rate = np.mean((sr_ret > sr_benchmark_ret) * 1)
                win_rate_0 = np.mean((sr_ret > 0) * 1)
                win_rate_weekly = np.mean((sr_ret_weekly > sr_benchmark_ret_weekly) * 1)
                win_rate_0_weekly = np.mean((sr_ret_weekly > 0) * 1)
                win_rate_monthly = np.mean((sr_ret_monthly > sr_benchmark_ret_monthly) * 1)
                win_rate_0_monthly = np.mean((sr_ret_monthly > 0) * 1)
                
                corr = sr_ret.corr(sr_benchmark_ret)    
                track_err = (sr_ret - sr_benchmark_ret).std(ddof=1) * np.sqrt(period_1y)
                info = excess_ret / track_err
                x = (sr_benchmark_ret - risk_free_per_period).dropna() .values
                y = (sr_ret - risk_free_per_period).dropna().values
                x = sm.add_constant(x)
                model = regression.linear_model.OLS(y,x).fit()
                alpha = model.params[0]
                beta = model.params[1]
                treynor = (annual_ret - risk_free_rate) / beta
                not_system_risk = sum((sr_ret - risk_free_per_period)**2) \
                                - alpha * sum(sr_ret - risk_free_per_period) \
                                - beta * sum((sr_ret - risk_free_per_period) * (sr_benchmark_ret - risk_free_per_period))
                not_system_risk = np.sqrt(not_system_risk / (len(sr_ret) - 2))
                alpha = alpha * period_1y
                df_ret = np.log(df).diff(1).iloc[1:]
                df_ret_up = df_ret[df_ret.index_ret > 0] + 1 
                up_cap_top = df_ret_up.fund_ret.product() ** (period_1y/len(df_ret_up)) - 1
                up_cap_bot = df_ret_up.index_ret.product() ** (period_1y/len(df_ret_up)) - 1
                up_capture = up_cap_top / up_cap_bot

                df_ret_down = df_ret[df_ret.index_ret < 0] + 1
                df_ret_down_length = max(1,len(df_ret_down))
                down_cap_top = df_ret_down.fund_ret.product() ** (period_1y/df_ret_down_length) - 1
                down_cap_bot = df_ret_down.index_ret.product() ** (period_1y/df_ret_down_length) - 1
                down_capture = down_cap_top / down_cap_bot

                df_ret['excess_ret'] = df_ret.fund_ret - df_ret.index_ret
                df_ret = df_ret.replace(np.Inf,np.nan).replace(-np.Inf,np.nan).dropna()
                if df_ret.shape[0] < 4:
                    excess_ret_cotinue = np.nan
                else:
                    mod = AutoReg(endog=df_ret.excess_ret.values, lags=1)
                    res = mod.fit()
                    excess_ret_cotinue = res.params[0]

                sr_excess_value = np.exp((sr_ret-sr_benchmark_ret).cumsum())
                _b_d = sr.loc[:sr_excess_value.index[0]].index[-2]
                sr_excess_value.loc[_b_d] = 1
                sr_excess_value = sr_excess_value.sort_index()
                mdd_part =  sr_excess_value / sr_excess_value.rolling(window=sr_excess_value.shape[0], min_periods=1).max()
                excess_mdd = 1 - mdd_part.min()
                
                
                if np.isnan(excess_mdd):
                    excess_mdd_date2 = None
                    excess_mdd_date1 = None
                    excess_mdd_lens = None
                else:
                    excess_mdd_date2 = mdd_part.idxmin()
                    excess_mdd_date1 = sr[:excess_mdd_date2].idxmax()
                    excess_mdd_lens = (excess_mdd_date2-excess_mdd_date1).days
                
                return {
                    'start_date':start_date,
                    'end_date':end_date,
                    'trade_year':trade_year,
                    'last_unit_nav':last_unit_nav,
                    'cumu_ret':cumu_ret,
                    'annual_ret':annual_ret,
                    'annual_vol':annual_vol,
                    'sharpe':sharpe,
                    'recent_year_ret':recent_year_ret,
                    'recent_1w_ret':recent_1w_ret,
                    'recent_1m_ret':recent_1m_ret,
                    'recent_4w_ret':recent_4w_ret,
                    'y_2020_ret':y_2020_ret,
                    'last_year_ret':last_year_ret,
                    'recent_3m_ret':recent_3m_ret,
                    'recent_6m_ret':recent_6m_ret,
                    'recent_1y_ret':recent_1y_ret,
                    'recent_3y_ret': recent_3y_ret,
                    'recent_5y_ret':recent_5y_ret,
                    'worst_3m_ret':worst_3m_ret,
                    'worst_6m_ret':worst_6m_ret,
                    'last_mv_diff':last_mv_diff,
                    'last_increase_rate':last_increase_rate,
                    'recent_drawdown':recent_drawdown,
                    'recent_mdd_date1':recent_mdd_date1,
                    'recent_mdd_lens':recent_mdd_lens,
                    'mdd':mdd,
                    'mdd_date1':mdd_date1,
                    'mdd_date2':mdd_date2,
                    'mdd_lens':mdd_lens,
                    'calmar':calmar,
                    'downside_risk':downside_risk,
                    'sortino':sortino,
                    'var':var,
                    'ervar':ervar,
                    'skew':skew_value,
                    'kurtosis':kurtosis_value,
                    'alpha_cl':alpha_cl,
                    'beta_cl':beta_cl,
                    'beta_bear':beta_bear,
                    'beta_bull':beta_bull,
                    'alpha_hm':alpha_hm,
                    'beta_hm':beta_hm,
                    'excess_ret':excess_ret,
                    'win_rate':win_rate,
                    'win_rate_0':win_rate_0,
                    'win_rate_weekly':win_rate_weekly,
                    'win_rate_0_weekly':win_rate_0_weekly,
                    'win_rate_monthly':win_rate_monthly,
                    'win_rate_0_monthly':win_rate_0_monthly,
                    'raise_month_num':raise_month_num,
                    'drop_month_num':drop_month_num,
                    'corr':corr,
                    'track_err':track_err,
                    'info':info,
                    'alpha':alpha,
                    'beta':beta,
                    'treynor':treynor,
                    'not_system_risk':not_system_risk,
                    'cvar':cvar,
                    'up_capture':up_capture,
                    'down_capture':down_capture,
                    'excess_ret_cotinue':excess_ret_cotinue,
                    'continue_value_m':continue_value_m,
                    'excess_mdd':excess_mdd,
                    'excess_mdd_date1':excess_mdd_date1,
                    'excess_mdd_date2':excess_mdd_date2,
                    'excess_mdd_lens':excess_mdd_lens,
                    'total_relative_ret':total_relative_ret,
                    'mdd_recover':mdd_recover,
                    'mdd_recover_date1':mdd_recover_date1,
                    'mdd_recover_date2':mdd_recover_date2,
                    'mdd_recover_lens':mdd_recover_lens,
                    'win_rate_0_period':win_rate_0_period,
                    'nav_density':nav_density,
                    'nav_type':nav_type,
                    'w1_begin_date':w1_begin_date,
                    'w1_end_date':w1_end_date,
                }
            else:
                return {
                    'start_date':start_date,
                    'end_date':end_date,
                    'trade_year':trade_year,
                    'last_unit_nav':last_unit_nav,
                    'cumu_ret':cumu_ret,
                    'annual_ret':annual_ret,
                    'annual_vol':annual_vol,
                    'sharpe':sharpe,
                    'recent_year_ret':recent_year_ret,
                    'recent_1w_ret':recent_1w_ret,
                    'recent_1m_ret':recent_1m_ret,
                    'recent_4w_ret':recent_4w_ret,
                    'y_2020_ret':y_2020_ret,
                    'last_year_ret':last_year_ret,
                    'recent_3m_ret':recent_3m_ret,
                    'recent_6m_ret':recent_6m_ret,
                    'recent_1y_ret':recent_1y_ret,
                    'recent_3y_ret': recent_3y_ret,
                    'recent_5y_ret':recent_5y_ret,
                    'worst_3m_ret':worst_3m_ret,
                    'worst_6m_ret':worst_6m_ret,
                    'last_mv_diff':last_mv_diff,
                    'last_increase_rate':last_increase_rate,
                    'recent_drawdown':recent_drawdown,
                    'recent_mdd_date1':recent_mdd_date1,
                    'recent_mdd_lens':recent_mdd_lens,
                    'raise_month_num':raise_month_num,
                    'drop_month_num':drop_month_num,
                    'mdd':mdd,
                    'mdd_date1':mdd_date1,
                    'mdd_date2':mdd_date2,
                    'mdd_lens':mdd_lens,
                    'calmar':calmar,
                    'downside_risk':downside_risk,
                    'sortino':sortino,
                    'var':var,
                    'cvar':cvar,
                    'ervar':ervar,
                    'skew':skew_value,
                    'kurtosis':kurtosis_value,
                    'mdd_recover':mdd_recover,
                    'mdd_recover_date1':mdd_recover_date1,
                    'mdd_recover_date2':mdd_recover_date2,
                    'mdd_recover_lens':mdd_recover_lens,
                    'win_rate_0_period':win_rate_0_period,
                    'nav_density':nav_density,
                    'nav_type':nav_type,
                    'w1_begin_date':w1_begin_date,
                    'w1_end_date':w1_end_date,
                }
        else:
            return {
                    'start_date':dates[0] if len(dates) > 0 else None,
                    'end_date':dates[-1] if len(dates) > 0 else None,
                    'trade_year':0,
                    'last_unit_nav':None,
                    'cumu_ret':0,
                    'annual_ret':0,
                    'annual_vol':0,
                    'sharpe':0,
                    'recent_year_ret':0,
                    'recent_1w_ret':0,
                    'recent_1m_ret':0,
                    'recent_4w_ret':0,
                    'y_2020_ret':0,
                    'last_year_ret':0,
                    'recent_3m_ret':0,
                    'recent_6m_ret':0,
                    'recent_1y_ret':0,
                    'recent_3y_ret': 0,
                    'recent_5y_ret':0,
                    'worst_3m_ret':0,
                    'worst_6m_ret':0,
                    'last_mv_diff':0,
                    'last_increase_rate':0,
                    'recent_drawdown':0,
                    'recent_mdd_date1':dates[0] if len(dates) > 0 else None,
                    'recent_mdd_lens':0,
                    'mdd':0,
                    'mdd_date1':dates[0] if len(dates) > 0 else None,
                    'mdd_date2':dates[0] if len(dates) > 0 else None,
                    'mdd_lens':0,
                    'calmar':0,
                    'downside_risk':0,
                    'sortino':0,
                    'var':0,
                    'cvar':0,
                    'ervar':0,
                    'skew':0,
                    'kurtosis':0,
                    'alpha_cl':0,
                    'beta_cl':0,
                    'beta_bear':0,
                    'beta_bull':0,
                    'alpha_hm':0,
                    'beta_hm':0,
                    'excess_ret':0,
                    'win_rate':0,
                    'win_rate_0':0,
                    'win_rate_weekly':0,
                    'win_rate_0_weekly':0,
                    'win_rate_monthly':0,
                    'win_rate_0_monthly':0,
                    'raise_month_num':0,
                    'drop_month_num':0,
                    'raise_month_num':0,
                    'drop_month_num':0,
                    'corr':0,
                    'track_err':0,
                    'info':0,
                    'alpha':0,
                    'beta':0,
                    'treynor':0,
                    'not_system_risk':0,
                    'cvar':0,
                    'up_capture':0,
                    'down_capture':0,
                    'excess_ret_cotinue':0,
                    'continue_value_m':0,
                    'excess_mdd':0,
                    'excess_mdd_date1':dates[0] if len(dates) > 0 else None,
                    'excess_mdd_date2':dates[0] if len(dates) > 0 else None,
                    'excess_mdd_lens':0,
                    'total_relative_ret':0,
                    'mdd_recover':0,
                    'mdd_recover_date1':dates[0] if len(dates) > 0 else None,
                    'mdd_recover_date2':dates[0] if len(dates) > 0 else None,
                    'mdd_recover_lens':0,
                    'win_rate_0_period':0,
                    'nav_density':0,
                    'nav_type':'nan',
                    'w1_begin_date':dates[0] if len(dates) > 0 else None,
                    'w1_end_date':dates[0] if len(dates) > 0 else None,
                }


    @staticmethod    
    def mdd_recover_analysis(values,dates):
        sr = pd.Series(values, index=dates).sort_index()
        if sr.empty:
            mdd = 0
            mdd_date1 = None
            mdd_date2 = None
            mdd_lens = 0
            return mdd, mdd_date1, mdd_date2, mdd_lens
        mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
        mdd = 1 - mdd_part.min()
        if mdd == 0:
            mdd_date1 = None
            mdd_date2 = None
            mdd_lens = 0
        else:
            mdd_date = mdd_part.idxmin()
            mdd_date1 = sr[:mdd_date].idxmax()
            sr_tmp = sr[mdd_date1:]
            recover_sr = sr_tmp[sr_tmp> sr[mdd_date1]]
            if recover_sr.empty:
                mdd_date2 = sr_tmp.index[-1]
            else: 
                mdd_date2 = sr_tmp[sr_tmp> sr[mdd_date1]].index[0]
            mdd_lens = sr.loc[mdd_date1:mdd_date2].shape[0]
        return mdd, mdd_date1, mdd_date2, mdd_lens

    @staticmethod
    def simulated_mv_calc(df,fee_dic,weight_dic):
        '''
        df structure: 
                    index   : datetime
                    column  : different assets mv
        i.g.

                	A	    B	    C	    D	    E
        2020-03-02	1.2047	1.0858	1.4002	1.0482	1.000
        2020-03-06	1.2156	1.0983	1.4211	1.0496	1.005
        
        mng_fee_dic: 
        i.g.            
            fee_dic = {'A':0.2,'B':0.2,'C':0.2,'D':0.2,'E':0.2}
        
        weight_dic: value or weight 
        i.g.
            weight_dic = {'A':1,'B':1,'C':1,'D':1,'E':1}

        '''
        # 归一化
        df_to_1 = df / df.iloc[0]
        # 盈利
        excess_ret = (df_to_1 - 1)
        # 盈 或 亏
        earn_con = (excess_ret > 0) * 1
        # 费
        pay_mng_fee = excess_ret * earn_con * pd.DataFrame([fee_dic]).values
        # 净值
        simu_nav = df_to_1
        simu_nav_vir = df_to_1 - pay_mng_fee
        # 真实总净值
        simu_nav_total = pd.DataFrame((simu_nav*pd.DataFrame([weight_dic]).values).sum(axis=1))
        # 虚拟总净值
        simu_nav_vir_total = pd.DataFrame((simu_nav_vir*pd.DataFrame([weight_dic]).values).sum(axis=1))
        simu_nav_total.columns = ['mv']
        simu_nav_vir_total.columns = ['mv']
        return simu_nav_vir_total, simu_nav_total

    @staticmethod
    def df_data_select(df, start_date, end_date, cols, resample_method, drop_col_title):
        # 组合净值分析时 净值选取
        # 设置 起始日 截止日 选取列名 对列共同名字是否去掉
        if end_date is None:
            end_date = datetime.date(2040,1,1)
        if start_date is None:
            start_date = datetime.date(2010,1,1)
        if cols is not None:
            df = df[cols]
        df = CalculatorBase.data_resample(df.loc[start_date:end_date], resample_method, drop_col_title)
        return df

    @staticmethod
    def data_resample(df, resample_method, drop_col_title=True):
        COL_TITLE = '融智'
        if resample_method == '1D':
            return df 
        if drop_col_title:
            cols = df.columns.tolist()
            cols = [i.replace(COL_TITLE,'') for i in cols]
            df.columns = cols
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(resample_method).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df.ffill()

    @staticmethod
    def proper_round(num, dec=0):
        dec = Decimal(10) ** -dec
        return float(Decimal(str(num)).quantize(dec, context=Context(rounding=ROUND_HALF_UP)))

    @staticmethod
    def stock_ipo_continue_up_limit(    dates:  pd.core.series.Series,
                                        values: pd.core.series.Series,
                                        ipo_price: float,
                                        list_date: datetime.date,
                                        ipo_date: datetime.date,
                                        trade_market: str,
                                        stock_name: str,
                                        stock_code: str):
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        assert len(dates) == len(values), 'date and port_values has different numbers'
        df = pd.DataFrame(values, index=dates).sort_index().dropna()
        col = 'price'
        df.columns = [col]
        rate = [1.44] + df.shape[0] * [1.1]
        df.loc[ipo_date,col] = ipo_price
        df = df.sort_index()
        if trade_market in ['上交所主板','深交所中小板']:
            df.loc[:,'limit_price'] = (df.price * rate).map(lambda x: CalculatorBase.proper_round(x,2)).shift(1)
            df.loc[:,'limit_con'] = (df.limit_price == df.price) * 1  
            limit_list = df.limit_con.iloc[1:].tolist()
            if 0 not in limit_list:
                ipo_lens = len(limit_list)
            else:
                ipo_lens = limit_list.index(0)
            num = CalculatorBase.STOCK_IPO_NUM_MARKET[trade_market]
            earning_per_stick = int(round((df.price.values[ipo_lens] - df.price.values[0]) * num))
        else:
            ipo_lens = np.nan
            earning_per_stick = np.nan
        res = {
            'stock_code':stock_code,
            'stock_name':stock_name,
            'trade_market':trade_market,
            'list_date':list_date,
            'ipo_price':ipo_price,
            'latest_price':df.price.values[-1],
            'latest_rate':df.price.pct_change(1).values[-1],
            'cumulate_rate':df.price.values[-1] / df.price.values[0] - 1,
            'first_date_rate':df.price.values[1] / df.price.values[0] - 1,
            'continue_limited_rise':ipo_lens,
            'earn_per_stick':earning_per_stick,
        }
        return res

    @staticmethod
    def conv_bond_ipo_result(   bond_id:str,
                                bond_name:str,
                                list_date:datetime.date,
                                ipo_price:float,
                                dates:pd.core.series.Series,
                                close_values:pd.core.series.Series,
                                open_values:pd.core.series.Series,
                                rate:float,
                                limit_num: Optional[str]):

        rate = rate / 100
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        assert len(dates) == len(close_values), 'date and port_values has different numbers'
        assert len(dates) == len(open_values), 'date and port_values has different numbers'
        df = pd.DataFrame({'close':close_values,'open':open_values}, index=dates).sort_index().dropna()
        if limit_num is not None:
            limit_num = int(limit_num.split('网上申购数量上限：')[1].split('手')[0])
        trade_market = bond_id.split('.')[1]
        new_price = df.close[-1]
        total_ret = df.close[-1] / ipo_price - 1
        first_open = df.open[0]
        first_date_rate = df.close[0] / 100 - 1
        earn_per_stick = (first_open - ipo_price) * 10
        if limit_num is not None:
            earn_per_account = earn_per_stick * limit_num * rate
        else:
            earn_per_account = np.nan
        res = {
            'bond_id':bond_id,
            'bond_name':bond_name,
            'list_date':list_date,
            'ipo_price':ipo_price,
            'trade_market':trade_market,
            'new_price':new_price,
            'total_ret':total_ret,
            'earn_per_account':earn_per_account,
            'earn_per_stick':earn_per_stick,
            'first_date_rate':first_date_rate,
            'first_open':first_open,
            'rate':rate,
            'limit_num':limit_num,
        }
        return res
    
    @staticmethod
    def data_resample_monthly_ret(df, rule='1M', min_count=15):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_weekly_ret(df, rule='1W', min_count=3):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_monthly_nav(df, rule='1M'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_weekly_nav(df, rule='1W'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def get_nav_type(values, dates):
        sr = pd.Series(values, index=dates).sort_index()
        tds = pd.to_datetime(sr.index)
        tds = [_.date() for _ in tds]
        sr.index = tds
        dts = BasicDataApi().get_trading_day_list(start_date=sr.index[0], end_date=sr.index[-1]).set_index('datetime')
        sr_daily = sr.reindex(dts.index)
        sr_week = CalculatorBase.data_resample_weekly_nav(df=sr_daily)
        sr_month = CalculatorBase.data_resample_monthly_nav(df=sr_daily)

        nav_density_daily = 1 - (sum(sr_daily.isnull() * 1) / sr_daily.shape[0])
        nav_density_weeky = 1 - (sum(sr_week.isnull() * 1) / sr_week.shape[0])
        nav_density_monthy = 1 - (sum(sr_month.isnull() * 1) / sr_month.shape[0])

        # 如果周度净值价格都在非交易日上，比如周日
        if sr.dropna().shape[0] > 0 and nav_density_monthy == 0:
            sr = pd.Series(values, index=dates).sort_index()
            sr_week = CalculatorBase.data_resample_weekly_nav(df=sr)
            sr_week = sr.reindex(dts.index)
            sr_week = CalculatorBase.data_resample_weekly_nav(df=sr)
            sr_month = CalculatorBase.data_resample_monthly_nav(df=sr_week)
            nav_density_weeky = 1 - (sum(sr_week.isnull() * 1) / sr_week.shape[0])
            nav_density_monthy = 1 - (sum(sr_month.isnull() * 1) / sr_month.shape[0])
            
        if nav_density_daily > 0.8:
            fund_nav_type = 'daily'
        elif nav_density_weeky > 0.8:
            fund_nav_type = 'weekly'
        else:
            fund_nav_type = 'monthly'
        return fund_nav_type



import datetime

from sqlalchemy import CHAR, Column, Integer, Index, BOOLEAN, text, TEXT, Enum
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, DATETIME, SMALLINT

from ...constant import StockFactorType, IndClassType, HoldingAssetType


class Base:
    _update_time = Column('_update_time', DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))  # 更新时间


class FOFBase:
    create_time = Column(DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP'))  # 创建时间
    update_time = Column(DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))  # 更新时间
    is_deleted = Column(BOOLEAN, nullable=False, server_default=text('FALSE'))  # 是否删除 默认不删除


# make this column at the end of every derived table
Base._update_time._creation_order = 9999
Base = declarative_base(cls=Base)

FOFBase.create_time._creation_order = 9997
FOFBase.update_time._creation_order = 9998
FOFBase.is_deleted._creation_order = 9999
FOFBase = declarative_base(cls=FOFBase)


class IndexVolatility(Base):
    '''指数波动率'''

    __tablename__ = 'index_volatility'

    index_id = Column(CHAR(20), primary_key=True) # 指数ID
    datetime = Column(DATE, primary_key=True) # 日期
    w1_vol = Column(DOUBLE(asdecimal=False)) # 近一周波动率
    m1_vol = Column(DOUBLE(asdecimal=False)) # 近一月波动率
    m3_vol = Column(DOUBLE(asdecimal=False)) # 近三月波动率
    m6_vol = Column(DOUBLE(asdecimal=False)) # 近半年波动率
    y1_vol = Column(DOUBLE(asdecimal=False)) # 近一年波动率
    y3_vol = Column(DOUBLE(asdecimal=False)) # 近三年波动率
    y5_vol = Column(DOUBLE(asdecimal=False)) # 近五年波动率
    y10_vol = Column(DOUBLE(asdecimal=False)) # 近十年波动率
    this_y_vol = Column(DOUBLE(asdecimal=False)) # 今年以来波动率
    this_s_vol = Column(DOUBLE(asdecimal=False)) # 本季度以来波动率
    cumulative_vol = Column(DOUBLE(asdecimal=False)) # 成立至今波动率

    __table_args__ = (
        Index('idx_index_volatility_datetime', 'datetime'),
    )


class IndexReturn(Base):
    '''指数收益率'''

    __tablename__ = 'index_return'

    index_id = Column(CHAR(20), primary_key=True) # 指数ID
    datetime = Column(DATE, primary_key=True) # 日期
    w1_ret = Column(DOUBLE(asdecimal=False)) # 近一周收益率
    m1_ret = Column(DOUBLE(asdecimal=False)) # 近一月收益率
    m3_ret = Column(DOUBLE(asdecimal=False)) # 近三月收益率
    m6_ret = Column(DOUBLE(asdecimal=False)) # 近半年收益率
    y1_ret = Column(DOUBLE(asdecimal=False)) # 近一年收益率
    y3_ret = Column(DOUBLE(asdecimal=False)) # 近三年收益率
    y5_ret = Column(DOUBLE(asdecimal=False)) # 近五年收益率
    y10_ret = Column(DOUBLE(asdecimal=False)) # 近十年收益率
    this_y_ret = Column(DOUBLE(asdecimal=False)) # 今年以来收益率
    this_s_ret = Column(DOUBLE(asdecimal=False)) # 本季度以来收益率
    cumulative_ret = Column(DOUBLE(asdecimal=False)) # 成立至今收益率

    __table_args__ = (
        Index('idx_index_return_datetime', 'datetime'),
    )


class FundAlpha(Base):
    '''基金超额收益'''

    __tablename__ = 'fund_alpha'

    track_err = Column(DOUBLE(asdecimal=False)) # 跟踪误差
    this_y_alpha = Column(DOUBLE(asdecimal=False)) # 今年以来超额收益
    cumulative_alpha = Column(DOUBLE(asdecimal=False)) # 成立以来超额收益
    w1_alpha = Column(DOUBLE(asdecimal=False)) # 近一周收益率
    m1_alpha = Column(DOUBLE(asdecimal=False)) # 近一月超额收益
    m3_alpha = Column(DOUBLE(asdecimal=False)) # 近三月超额收益
    m6_alpha = Column(DOUBLE(asdecimal=False)) # 近半年超额收益
    y1_alpha = Column(DOUBLE(asdecimal=False)) # 近一年超额收益
    y3_alpha = Column(DOUBLE(asdecimal=False)) # 近三年超额收益
    y5_alpha = Column(DOUBLE(asdecimal=False)) # 近五年超额收益
    y10_alpha = Column(DOUBLE(asdecimal=False)) # 近十年超额收益
    fund_id = Column(CHAR(16), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期

    __table_args__ = (
        Index('idx_fund_alpha_datetime', 'datetime'),
    )


class IndexValuationLongTerm(Base):
    '''指数估值'''

    __tablename__ = 'index_valuation_long_term'

    index_id = Column(CHAR(20), primary_key=True) #指数ID
    datetime = Column(DATE, primary_key=True) # 日期

    pb_mrq = Column(DOUBLE(asdecimal=False)) # 市净率-MRQ
    pe_ttm = Column(DOUBLE(asdecimal=False)) # 市盈率-MMT
    pe_ttm_nn = Column(DOUBLE(asdecimal=False)) # 市盈率PE(TTM)(剔除负值)
    roe = Column(DOUBLE(asdecimal=False)) # 净资产收益率
    ps_ttm = Column(DOUBLE(asdecimal=False)) # 市销率—MMT
    dy = Column(DOUBLE(asdecimal=False)) # 股息率
    pcf_ttm = Column(DOUBLE(asdecimal=False)) # 市现率-MMT
    peg_ttm = Column(DOUBLE(asdecimal=False)) # 市盈率相对盈利增长比率-MMT
    pe_pct = Column(DOUBLE(asdecimal=False)) # PE百分位(10/5年)
    pb_pct = Column(DOUBLE(asdecimal=False)) # PB百分位(10/5年)
    roe_pct = Column(DOUBLE(asdecimal=False)) # ROE百分位(10/5年)
    pe_pct_30 = Column(DOUBLE(asdecimal=False)) # PE百分位(10/5年) 30 分位
    pe_pct_70 = Column(DOUBLE(asdecimal=False)) # PE百分位(10/5年) 70 分位
    pb_pct_30 = Column(DOUBLE(asdecimal=False)) # PB百分位(10/5年) 30 分位
    pb_pct_70 = Column(DOUBLE(asdecimal=False)) # PB百分位(10/5年) 70 分位
    roe_pct_30 = Column(DOUBLE(asdecimal=False)) # ROE百分位(10/5年) 30 分位
    roe_pct_70 = Column(DOUBLE(asdecimal=False)) # ROE百分位(10/5年) 70 分位
    val_score = Column(DOUBLE(asdecimal=False)) # 估值评分(10/5年)
    eps_ttm = Column(DOUBLE(asdecimal=False)) # 每股收益—MMT
    ps_pct = Column(DOUBLE(asdecimal=False)) # PS百分位(10/5年)
    est_peg = Column(DOUBLE(asdecimal=False)) # 预测peg
    pe_pct_5_3 = Column(DOUBLE(asdecimal=False)) # PE百分位
    pb_pct_5_3 = Column(DOUBLE(asdecimal=False)) # PB百分位
    ps_pct_5_3 = Column(DOUBLE(asdecimal=False)) # PS百分位
    roe_pct_5_3 = Column(DOUBLE(asdecimal=False)) # ROE百分位
    mv = Column(DOUBLE(asdecimal=False)) # 总市值
    liqmv = Column(DOUBLE(asdecimal=False)) # 流通市值
    freeliqmv = Column(DOUBLE(asdecimal=False)) # 自由流通市值

    __table_args__ = (
        Index('idx_index_valuation_long_term_datetime', 'datetime'),
    )

class FundIndicator(Base):
    '''基金评价指标'''

    __tablename__ = 'fund_indicator'
    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    beta = Column(DOUBLE(asdecimal=False)) # 风险指数
    alpha = Column(DOUBLE(asdecimal=False)) # 投资回报
    track_err = Column(DOUBLE(asdecimal=False)) # 跟踪误差
    timespan = Column(DOUBLE(asdecimal=False)) # 历史数据跨度(年)
    fee_rate = Column(DOUBLE(asdecimal=False)) # 费率
    info_ratio = Column(DOUBLE(asdecimal=False)) # 信息比率
    treynor = Column(DOUBLE(asdecimal=False)) # 特雷诺比率
    mdd = Column(DOUBLE(asdecimal=False)) #净值最大回撤
    down_risk = Column(DOUBLE(asdecimal=False)) #下行风险
    ret_over_period = Column(DOUBLE(asdecimal=False)) #区间收益率
    annual_avg_daily_ret = Column(DOUBLE(asdecimal=False)) #年化日均收益
    annual_vol = Column(DOUBLE(asdecimal=False)) #年化波动率
    annual_ret = Column(DOUBLE(asdecimal=False)) #年化收益
    m_square = Column(DOUBLE(asdecimal=False)) #M平方测度 风险调整收益指标
    time_ret = Column(DOUBLE(asdecimal=False)) #择时收益
    var = Column(DOUBLE(asdecimal=False)) #资产在险值
    r_square = Column(DOUBLE(asdecimal=False)) #决定系数R方
    sharpe = Column(DOUBLE(asdecimal=False)) #夏普率
    year_length = Column(DOUBLE(asdecimal=False)) #成立年限
    alpha_bond = Column(DOUBLE(asdecimal=False)) #alpha 债
    beta_bond = Column(DOUBLE(asdecimal=False)) #beta 债
    alpha_hs300 = Column(DOUBLE(asdecimal=False)) #alpha hs300
    beta_hs300 = Column(DOUBLE(asdecimal=False)) #beta hs300

    __table_args__ = (
        Index('idx_fund_indicator_datetime', 'datetime'),
    )


class FundIndicatorWeekly(Base):
    '''基金评价指标'''

    __tablename__ = 'fund_indicator_weekly'
    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    beta_w = Column(DOUBLE(asdecimal=False)) # 风险指数
    alpha_w = Column(DOUBLE(asdecimal=False)) # 投资回报
    track_err_w = Column(DOUBLE(asdecimal=False)) # 跟踪误差

    __table_args__ = (
        Index('idx_fund_indicator_datetime_weekly', 'datetime'),
    )

class FundIndicatorMonthly(Base):
    '''基金评价指标（月度）'''

    __tablename__ = 'fund_indicator_monthly'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金ID
    datetime = Column(DATE, primary_key=True)  # 日期
    beta_m = Column(DOUBLE(asdecimal=False), nullable=False)  # beta
    sharpe_ratio_m = Column(DOUBLE(asdecimal=False))  # 夏普比率
    treynor_ratio_m = Column(DOUBLE(asdecimal=False))  # 特雷诺比率
    information_ratio_m = Column(DOUBLE(asdecimal=False))  # 信息比率
    jensen_alpha_m = Column(DOUBLE(asdecimal=False), nullable=False)  # 詹森指数
    calmar_ratio_m = Column(DOUBLE(asdecimal=False))  # 卡玛比率

    __table_args__ = (
        Index('idx_fund_indicator_monthly_datetime_monthly', 'datetime'),
    )


class FundIndicatorAnnual(Base):
    '''基金年度指标'''

    __tablename__ = 'fund_indicator_annual'

    fund_id = Column(CHAR(16), primary_key=True)  # 基金ID
    datetime = Column(DATE, primary_key=True)  # 日期
    beta_annual = Column(DOUBLE(asdecimal=False))  # beta
    alpha_annual = Column(DOUBLE(asdecimal=False))  # alpha
    excess_return_annual = Column(DOUBLE(asdecimal=False))  # 超额收益

    __table_args__ = (
        Index('idx_fund_indicator_datetime_annual', 'datetime'),
    )


class FundIndicatorGroup(Base):
    '''基金评价指标'''

    __tablename__ = 'fund_indicator_group'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金ID
    datetime = Column(DATE, primary_key=True)  # 日期
    data_cycle = Column(CHAR(3), primary_key=True)  # 使用多长时间的数据来计算
    scale = Column(DOUBLE(asdecimal=False))  # 规模
    beta = Column(DOUBLE(asdecimal=False))
    alpha = Column(DOUBLE(asdecimal=False))
    track_err = Column(DOUBLE(asdecimal=False))  # 跟踪误差
    info_ratio = Column(DOUBLE(asdecimal=False))  # 信息比率
    info_ratio_hs300 = Column(DOUBLE(asdecimal=False))  # 相对沪深300的信息比率
    treynor = Column(DOUBLE(asdecimal=False))  # 特雷诺比率
    mdd = Column(DOUBLE(asdecimal=False))  # 净值最大回撤
    mdd_len = Column(DOUBLE(asdecimal=False))  # 最大回撤持续时间
    winning_rate = Column(DOUBLE(asdecimal=False))  # 和mmf的相对胜率
    VaR = Column(DOUBLE(asdecimal=False))  # Value at Risk
    CVaR = Column(DOUBLE(asdecimal=False))  # Expected Shortfall
    ins_holds = Column(DOUBLE(asdecimal=False))  # 机构持有者占比
    hold_num = Column(DOUBLE(asdecimal=False))  # 持有人数量
    leverage = Column(DOUBLE(asdecimal=False))  # 杠杆
    ptm = Column(DOUBLE(asdecimal=False))  # 到期期限
    annual_vol = Column(DOUBLE(asdecimal=False))  # 年化波动率
    annual_ret = Column(DOUBLE(asdecimal=False))  # 年化收益
    sharpe = Column(DOUBLE(asdecimal=False))
    calma_ratio = Column(DOUBLE(asdecimal=False))
    stutzer = Column(DOUBLE(asdecimal=False))
    continue_regress_v = Column(DOUBLE(asdecimal=False))  # 业绩持续性回归系数
    continue_regress_t = Column(DOUBLE(asdecimal=False))  # 业绩持续性回归T值
    continue_stats_v = Column(DOUBLE(asdecimal=False))  # 业绩持续性回归计算值
    stock_cl_alpha = Column(DOUBLE(asdecimal=False))  # cl alpha
    stock_cl_alpha_t = Column(DOUBLE(asdecimal=False))  # cl alpha t值
    stock_cl_alpha_p = Column(DOUBLE(asdecimal=False))  # cl alpha p值
    stock_cl_beta = Column(DOUBLE(asdecimal=False))  # cl beta
    stock_cl_beta_t = Column(DOUBLE(asdecimal=False))  # cl beta t值
    stock_cl_beta_p = Column(DOUBLE(asdecimal=False))  # cl beta p值
    stock_tm_alpha = Column(DOUBLE(asdecimal=False))  # tm alpha
    stock_tm_alpha_t = Column(DOUBLE(asdecimal=False))  # tm alpha t值
    stock_tm_alpha_p = Column(DOUBLE(asdecimal=False))  # tm alpha p值
    stock_tm_beta = Column(DOUBLE(asdecimal=False))  # tm beta
    stock_tm_beta_t = Column(DOUBLE(asdecimal=False))  # tm beta t值
    stock_tm_beta_p = Column(DOUBLE(asdecimal=False))  # tm beta p值
    stock_hm_alpha = Column(DOUBLE(asdecimal=False))  # hm alpha
    stock_hm_alpha_t = Column(DOUBLE(asdecimal=False))  # hm alpha t值
    stock_hm_alpha_p = Column(DOUBLE(asdecimal=False))  # hm alpha p值
    stock_hm_beta = Column(DOUBLE(asdecimal=False))  # hm beta
    stock_hm_beta_t = Column(DOUBLE(asdecimal=False))  # hm beta t值
    stock_hm_beta_p = Column(DOUBLE(asdecimal=False))  # hm beta p值
    abnormal_reason = Column(CHAR(60)) # 未计算因子 异常原因
    __table_args__ = (
        Index('idx_fund_indicator_group_datetime_data_cycle', 'datetime', 'data_cycle'),
    )

class FundScoreExtended(Base):
    '''新基金评分v1'''

    __tablename__ = 'fund_score_extended'

    fund_id = Column(CHAR(16), primary_key=True)  # 基金代码
    datetime = Column(DATE, primary_key=True)  # 日期
    data_cycle = Column(CHAR(3), primary_key=True)  # 使用多长时间的数据来计算
    wind_class_1 = Column(CHAR(16))  # Wind基金一级类型
    ret = Column(DOUBLE(asdecimal=False))  # ret
    vol = Column(DOUBLE(asdecimal=False))  # volatility
    sr = Column(DOUBLE(asdecimal=False))  # sharpe ratio
    ir = Column(DOUBLE(asdecimal=False))  # information ratio
    beta = Column(DOUBLE(asdecimal=False))  # beta
    alpha = Column(DOUBLE(asdecimal=False))  # alpha
    treynor_ratio = Column(DOUBLE(asdecimal=False))  # treynor_ratio
    mdd = Column(DOUBLE(asdecimal=False))  # max drawdown
    timing = Column(DOUBLE(asdecimal=False))  # timing
    ret_score = Column(DOUBLE(asdecimal=False))  # score of ret
    vol_score = Column(DOUBLE(asdecimal=False))  # score of volatility
    sr_score = Column(DOUBLE(asdecimal=False))  # score of sharpe ratio
    ir_score = Column(DOUBLE(asdecimal=False))  # score of information ratio
    beta_score = Column(DOUBLE(asdecimal=False))  # score of beta
    alpha_score = Column(DOUBLE(asdecimal=False))  # score of alpha
    treynor_ratio_score = Column(DOUBLE(asdecimal=False))  # score of treynor_ratio
    mdd_score = Column(DOUBLE(asdecimal=False))  # score of max drawdown
    return_score = Column(DOUBLE(asdecimal=False))  # score of return
    robust_score = Column(DOUBLE(asdecimal=False))  # score of robust
    timing_score = Column(DOUBLE(asdecimal=False))  # score of timing
    selection_score = Column(DOUBLE(asdecimal=False))  # score of selection
    allocation_score = Column(DOUBLE(asdecimal=False))  # score of allocation
    interaction_score = Column(DOUBLE(asdecimal=False))  # score of interaction
    return_rank = Column(DOUBLE(asdecimal=False))  # rank of return
    robust_rank = Column(DOUBLE(asdecimal=False))  # rank of robust
    timing_rank = Column(DOUBLE(asdecimal=False))  # rank of timing
    selection_rank = Column(DOUBLE(asdecimal=False))  # rank of selection
    allocation_rank = Column(DOUBLE(asdecimal=False))  # rank of allocation
    interaction_rank = Column(DOUBLE(asdecimal=False))  # rank of interaction
    total_score = Column(DOUBLE(asdecimal=False), nullable=False)  # score of total
    total_rank = Column(DOUBLE(asdecimal=False))  # rank of total

    __table_args__ = (
        Index('idx_fund_score_extended_datetime_data_cycle', 'datetime', 'data_cycle'),
    )

class FundScoreNew(Base):
    '''新基金评分v2'''

    __tablename__ = 'fund_score_new'

    fund_id = Column(CHAR(16), primary_key=True)  # 基金代码
    datetime = Column(DATE, primary_key=True)  # 日期
    data_cycle = Column(CHAR(3), primary_key=True)  # 使用多长时间的数据来计算
    wind_class_1 = Column(CHAR(16))  # Wind基金一级类型
    return_score = Column(DOUBLE(asdecimal=False))  # 被动指数型、被动指数型债券(均为二级分类)基金为跟踪标的能力，其他为收益能力
    robust_score = Column(DOUBLE(asdecimal=False))  # 稳定性能力
    risk_score = Column(DOUBLE(asdecimal=False))  # 抗风险能力
    timing_score = Column(DOUBLE(asdecimal=False))  # 被动指数型、被动指数型债券(均为二级分类)基金和货基(一二级分类相同)为管理规模能力，其他为择时能力
    selection_score = Column(DOUBLE(asdecimal=False))  # 被动指数型、被动指数型债券(均为二级分类)基金和货基(一二级分类相同)为机构观点，其他为选证能力
    team_score = Column(DOUBLE(asdecimal=False))  # 基金公司团队能力
    total_score = Column(DOUBLE(asdecimal=False), nullable=False)  # 总分
    abnormal_reason = Column(CHAR(60)) # 未计算因子 异常原因

    __table_args__ = (
        Index('idx_fund_score_new_data_cycle', 'data_cycle'),
    )

class AssetAllocationInfo(Base):
    '''十档大类资产组合'''

    __tablename__ = 'asset_allocation_info'
    allocation_id = Column(Integer, primary_key=True) # 组合id
    version = Column(Integer, primary_key=True) # 版本号

    hs300 =  Column(DOUBLE(asdecimal=False)) # 沪深300权重
    csi500 =  Column(DOUBLE(asdecimal=False)) # 中证500权重
    gem =  Column(DOUBLE(asdecimal=False)) # 创业板权重
    sp500rmb =  Column(DOUBLE(asdecimal=False)) # 标普500人民币权重
    national_debt = Column(DOUBLE(asdecimal=False)) # 国债权重
    gold =  Column(DOUBLE(asdecimal=False)) # 黄金权重
    credit_debt =  Column(DOUBLE(asdecimal=False)) # 信用债权重
    dax30rmb = Column(DOUBLE(asdecimal=False)) # 德国dax30权重
    real_state = Column(DOUBLE(asdecimal=False)) # 房地产权重 拼错了
    oil = Column(DOUBLE(asdecimal=False)) # 原油权重
    n225rmb = Column(DOUBLE(asdecimal=False)) # 日经225权重
    cash = Column(DOUBLE(asdecimal=False)) # 现金权重
    mdd = Column(DOUBLE(asdecimal=False)) # 最大回撤
    annual_ret = Column(DOUBLE(asdecimal=False)) # 年化收益
    sharpe = Column(DOUBLE(asdecimal=False)) # 夏普率
    recent_y5_ret = Column(DOUBLE(asdecimal=False)) # 最近五年收益
    annual_vol = Column(DOUBLE(asdecimal=False)) # 年化波动率
    mdd_d1 = Column(DATE) # 最大回撤开始时间
    mdd_d2 = Column(DATE) # 最大回撤结束时间
    start_date = Column(DATE) # 回测开始时间
    end_date = Column(DATE) # 回测结束时间

class StyleAnalysisFactorReturn(Base):
    '''风格分析因子收益'''

    __tablename__ = 'style_analysis_factor_return'

    universe_index = Column(CHAR(20), primary_key=True)  # universe, 全市场股票时值为all
    datetime = Column(DATE, primary_key=True)  # 日期
    latest_size = Column(DOUBLE(asdecimal=False))  # 规模
    bp = Column(DOUBLE(asdecimal=False))  # 价值
    short_term_momentum = Column(DOUBLE(asdecimal=False))  # 短期动量
    long_term_momentum = Column(DOUBLE(asdecimal=False))  # 长期动量
    high_low = Column(DOUBLE(asdecimal=False))  # 波动率
    const = Column(DOUBLE(asdecimal=False))  # 常数项

class BarraCNE5FactorReturn(Base):
    '''Barra CNE5因子收益'''

    __tablename__ = 'barra_cne5_factor_return'

    datetime = Column(DATE, primary_key=True)  # 日期
    market = Column(DOUBLE(asdecimal=False))
    leverage = Column(DOUBLE(asdecimal=False))
    cap = Column(DOUBLE(asdecimal=False))
    momentum = Column(DOUBLE(asdecimal=False))
    bp = Column(DOUBLE(asdecimal=False))
    earning = Column(DOUBLE(asdecimal=False))
    liquidity = Column(DOUBLE(asdecimal=False))
    growth = Column(DOUBLE(asdecimal=False))
    vol = Column(DOUBLE(asdecimal=False))

class AllocationDistribution(Base):
    '''有效前沿下各mdd最优策略'''

    __tablename__ = 'allocation_distribution'

    ef_id = Column(CHAR(8), primary_key=True)  # 有效前沿id
    version = Column(Integer, primary_key=True) # 版本号
    annual_ret = Column(DOUBLE(asdecimal=False)) #年化收益
    csi500 = Column(DOUBLE(asdecimal=False)) # 中证500权重
    gem = Column(DOUBLE(asdecimal=False)) #创业板权重
    gold = Column(DOUBLE(asdecimal=False)) #黄金权重
    hs300 = Column(DOUBLE(asdecimal=False)) #沪深300权重
    mdd = Column(DOUBLE(asdecimal=False)) # 最大回撤
    mdd_up_limit = Column(DOUBLE(asdecimal=False)) #最大回测档位上限
    mmf = Column(DOUBLE(asdecimal=False)) # 货币基金权重
    national_debt = Column(DOUBLE(asdecimal=False)) #国债权重
    sp500rmb = Column(DOUBLE(asdecimal=False)) #标普500权重
    cash = Column(DOUBLE(asdecimal=False)) #现金权重

class FundAIC(Base):
    '''基金定投收益收集'''

    __tablename__ = 'fund_aic'

    fund_id = Column(CHAR(20), primary_key=True)  # 基金ID
    datetime = Column(CHAR(20), primary_key=True)  # 时间
    y1_ret = Column(DOUBLE(asdecimal=False))  # 普通定投一年收益
    y3_ret = Column(DOUBLE(asdecimal=False))  # 普通定投三年收益
    y5_ret = Column(DOUBLE(asdecimal=False))  # 普通定投五年收益
    intel_y1_ret = Column(DOUBLE(asdecimal=False))  # 智能定投一年收益
    intel_y3_ret = Column(DOUBLE(asdecimal=False))  # 智能定投三年收益
    intel_y5_ret = Column(DOUBLE(asdecimal=False))  # 智能定投五年收益

class StyleBox(Base):
    '''基金风格九宫格评分'''

    __tablename__ = 'fund_style_box'

    fund_id = Column(CHAR(20), primary_key=True)  # 基金ID
    datetime = Column(DATE, primary_key=True)  # 时间
    x = Column(DOUBLE(asdecimal=False))  # 价值-平衡-成长维度
    y = Column(DOUBLE(asdecimal=False))  # 小盘-中盘-大盘维度

    __table_args__ = (
        Index('idx_fund_style_box_datetime', 'datetime'),
    )

class StyleReg(Base):
    '''Style 因子收益'''

    __tablename__ = 'style_reg_factor_return'

    datetime = Column(DATE, primary_key=True)  # 日期
    large_g= Column(DOUBLE(asdecimal=False))
    large_v = Column(DOUBLE(asdecimal=False))
    mid_g = Column(DOUBLE(asdecimal=False))
    mid_v = Column(DOUBLE(asdecimal=False))
    small_g = Column(DOUBLE(asdecimal=False))
    small_v = Column(DOUBLE(asdecimal=False))

class FundManagerIndex(Base):
    '''基金经理指数'''

    __tablename__ = 'fund_manager_index'

    manager_id = Column(CHAR(16), primary_key=True)  # 基金经理ID
    datetime = Column(DATE, primary_key=True)  # 日期
    fund_type = Column(CHAR(8), primary_key=True)  # 所管理基金类型
    manager_index = Column(DOUBLE(asdecimal=False))  # 基金经理指数

class FundManagerScore(Base):
    '''基金经理评分以及因子'''

    __tablename__ = 'fund_manager_score'

    manager_id = Column(CHAR(16), primary_key=True)  # 基金经理ID
    datetime = Column(DATE, primary_key=True)  # 日期
    fund_type = Column(CHAR(8), primary_key=True)  # 所管理基金类型
    trade_year = Column(DOUBLE(asdecimal=False))
    annualized_ret = Column(DOUBLE(asdecimal=False))
    annualized_vol = Column(DOUBLE(asdecimal=False))
    sharpe = Column(DOUBLE(asdecimal=False))
    mdd = Column(DOUBLE(asdecimal=False))
    alpha = Column(DOUBLE(asdecimal=False))
    beta = Column(DOUBLE(asdecimal=False))
    track_err = Column(DOUBLE(asdecimal=False))
    fund_size = Column(DOUBLE(asdecimal=False))
    model_alpha = Column(DOUBLE(asdecimal=False))
    model_beta = Column(DOUBLE(asdecimal=False))
    now_comp_year = Column(DOUBLE(asdecimal=False)) # 在当前基金公司时间
    year_ret_std =  Column(DOUBLE(asdecimal=False)) # 每年收益波动率
    score = Column(DOUBLE(asdecimal=False))

    __table_args__ = (
        Index('idx_fund_manager_score_datetime', 'datetime'),
    )

class FundManagerInfo(Base):
    '''基金经理信息'''
    __tablename__ = 'manager_info'

    mng_id = Column(CHAR(10), primary_key=True) # 基金经理id
    start_date = Column(DATE, primary_key=True) # 基金经理任职时间
    fund_id = Column(CHAR(10), primary_key=True) # 代码
    mng_name = Column(CHAR(25)) # 基金经理
    gender = Column(CHAR(2)) # 性别
    birth_year = Column(CHAR(4)) # 出生年份
    education = Column(CHAR(4)) # 学历
    nationality = Column(CHAR(10)) # 国籍
    resume = Column(TEXT) # 简历
    end_date = Column(DATE) # 基金经理离职时间
    company_id = Column(CHAR(64)) # 基金公司
    profit = Column(DOUBLE)     # 任期基金收益

class FundManagerFundRank(Base):
    '''基金经理基金排名'''
    __tablename__ = 'manager_fund_rank'

    mng_id = Column(CHAR(10), primary_key=True) # 基金经理id
    datetime = Column(DATE, primary_key=True) # 时间
    fund_list = Column(TEXT)

class ManagerFundDetail(Base):
    '''基金经理基金排名'''
    __tablename__ = 'manager_fund_detail'

    mng_id = Column(CHAR(10), primary_key=True) # 基金经理id
    datetime = Column(DATE, primary_key=True) # 时间
    stock_fund_list = Column(TEXT)
    # bond_fund_list = Column(TEXT)
    # mmf_fund_list = Column(TEXT)
    # qdii_fund_list = Column(TEXT)
    # mix_fund_list = Column(TEXT)

class NewShareFundRank(Base):
    """打新基金榜单"""
    __tablename__ = 'fund_rank_new_share'

    fund_id = Column(CHAR(10), primary_key=True)        # 基金ID
    final_score = Column(DOUBLE(asdecimal=False))       # 最终分数


class AbsReturnFundRank(Base):
    """绝对收益基金榜单"""
    __tablename__ = 'fund_rank_abs_return'

    fund_id = Column(CHAR(10), primary_key=True)        # 基金ID
    final_score = Column(DOUBLE(asdecimal=False))       # 最终分数


class ConvertibleBondFundRank(Base):
    """可转债基金榜单"""
    __tablename__ = 'fund_rank_convertible_bonds'

    fund_id = Column(CHAR(10), primary_key=True)        # 基金ID
    final_score = Column(DOUBLE(asdecimal=False))       # 最终分数


class ThirdPartyPortfolioInfo(Base):
    '''组合基本信息表'''

    __tablename__ = 'third_party_portfolio_info'

    po_id = Column(CHAR(16), primary_key=True)  # 组合ID
    po_src = Column(SMALLINT, nullable=False)  # 组合来源 详见constant.SourceType
    po_name = Column(TEXT, nullable=False)  # 组合名称
    po_desc = Column(TEXT)  # 组合描述
    manager_name = Column(TEXT) # 主理人名称
    rich_desc = Column(TEXT)  # 详细描述
    benchmarks = Column(TEXT) # 业绩比较基准
    start_date = Column(DATE, nullable=False)  # 组合开始日期
    risk_level = Column(CHAR(6))  # 风险级别
    can_aip = Column(BOOLEAN)  # 是否可定投


class ThirdPartyPortfolioTrade(Base):
    '''组合调仓记录表'''

    __tablename__ = 'third_party_portfolio_trade'

    po_id = Column(CHAR(16), primary_key=True)  # 组合ID
    datetime = Column(DATE, primary_key=True)  # 日期
    po_src = Column(SMALLINT, nullable=False)  # 组合来源 详见constant.SourceType
    details = Column(TEXT, nullable=False)  # 调仓详情
    comment = Column(TEXT)


class ThirdPartyPortfolioPositionLatest(Base):
    '''组合最新持仓表'''

    __tablename__ = 'third_party_portfolio_position_latest'

    po_id = Column(CHAR(16), primary_key=True)  # 组合ID
    datetime = Column(DATE, primary_key=True)  # 日期
    po_src = Column(SMALLINT, nullable=False)  # 组合来源 详见constant.SourceType
    position = Column(TEXT, nullable=False)  # 持仓

    __table_args__ = (
        Index('idx_third_party_portfolio_position_latest_po_src', 'po_src'),
    )

class ThirdPartyPortfolioIndicator(Base):
    '''组合最新因子表'''

    __tablename__ = 'third_party_portfolio_indicator'

    po_id = Column(CHAR(16), primary_key=True)  # 组合ID
    datetime = Column(DATE, primary_key=True)  # 日期
    po_src = Column(SMALLINT, nullable=False)  # 组合来源 详见constant.SourceType
    nav = Column(DOUBLE(asdecimal=False), nullable=False)  # 净值（累计复权净值计算得出）
    sharpe_ratio = Column(DOUBLE(asdecimal=False))  # 夏普比率
    mdd = Column(DOUBLE(asdecimal=False), nullable=False)  # 最大回撤
    annual_compounded_ret = Column(DOUBLE(asdecimal=False), nullable=False)  # 年化复合收益率
    daily_ret = Column(DOUBLE(asdecimal=False), nullable=False)  # 最近一天收益率
    weekly_ret = Column(DOUBLE(asdecimal=False))  # 最近一周收益率
    monthly_ret = Column(DOUBLE(asdecimal=False))  # 最近一月收益率
    quarterly_ret = Column(DOUBLE(asdecimal=False))  # 最近一季收益率
    half_yearly_ret = Column(DOUBLE(asdecimal=False))  # 最近半年收益率
    yearly_ret = Column(DOUBLE(asdecimal=False))  # 最近一年收益率
    from_setup_ret = Column(DOUBLE(asdecimal=False), nullable=False)  # 成立以来收益率
    vol_by_day = Column(DOUBLE(asdecimal=False))  # 年化日波动率
    vol_by_week = Column(DOUBLE(asdecimal=False))  # 年化周波动率
    vol_by_month = Column(DOUBLE(asdecimal=False))  # 年化月波动率
    annual_ret = Column(DOUBLE(asdecimal=False), nullable=False)  # 年化收益
    mdd_d1 = Column(DATE, nullable=False)  # 回撤区间高点日期
    mdd_d2 = Column(DATE, nullable=False)  # 回撤区间低点日期
    total_commission = Column(DOUBLE(asdecimal=False), nullable=False)  # 总手续费
    rebalance_times = Column(DOUBLE(asdecimal=False), nullable=False)  # 调仓次数

class FundIndicatorIPOConv(Base):
    '''基金评价指标'''

    __tablename__ = 'fund_indicator_ipo_conv'
    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    beta = Column(DOUBLE(asdecimal=False)) # 风险指数
    alpha = Column(DOUBLE(asdecimal=False)) # 投资回报
    track_err = Column(DOUBLE(asdecimal=False)) # 跟踪误差
    timespan = Column(DOUBLE(asdecimal=False)) # 历史数据跨度(年)
    fee_rate = Column(DOUBLE(asdecimal=False)) # 费率
    info_ratio = Column(DOUBLE(asdecimal=False)) # 信息比率
    treynor = Column(DOUBLE(asdecimal=False)) # 特雷诺比率
    mdd = Column(DOUBLE(asdecimal=False)) #净值最大回撤
    down_risk = Column(DOUBLE(asdecimal=False)) #下行风险
    ret_over_period = Column(DOUBLE(asdecimal=False)) #区间收益率
    annual_avg_daily_ret = Column(DOUBLE(asdecimal=False)) #年化日均收益
    annual_vol = Column(DOUBLE(asdecimal=False)) #年化波动率
    annual_ret = Column(DOUBLE(asdecimal=False)) #年化收益
    m_square = Column(DOUBLE(asdecimal=False)) #M平方测度 风险调整收益指标
    time_ret = Column(DOUBLE(asdecimal=False)) #择时收益
    var = Column(DOUBLE(asdecimal=False)) #资产在险值
    r_square = Column(DOUBLE(asdecimal=False)) #决定系数R方
    sharpe = Column(DOUBLE(asdecimal=False)) #夏普率
    year_length = Column(DOUBLE(asdecimal=False)) #成立年限

    __table_args__ = (
        Index('idx_fund_indicator_ipo_conv_datetime', 'datetime'),
    )

class TopManagerFunds(Base):
    '''基金经理选基'''

    __tablename__ = 'top_manager_funds'
    datetime = Column(DATE, primary_key=True) # 日期
    top_1 = Column(CHAR(10))  # 第一名基金
    top_2 = Column(CHAR(10))  # 第二名基金
    top_3 = Column(CHAR(10))  # 第三名基金
    top_4 = Column(CHAR(10))  # 第四名基金
    top_5 = Column(CHAR(10))  # 第五名基金
    top_6 = Column(CHAR(10))  # 第六名基金
    top_7 = Column(CHAR(10))  # 第七名基金
    top_8 = Column(CHAR(10))  # 第八名基金
    top_9 = Column(CHAR(10))  # 第九名基金
    top_10 = Column(CHAR(10)) # 第十名基金

class StockFactorInfo(Base):
    '''因子信息表'''

    __tablename__ = 'stock_factor_info'

    f_id = Column(CHAR(16), primary_key=True)  # 因子ID
    f_type = Column(Enum(StockFactorType), nullable=False)  # 因子所属分类
    f_name = Column(TEXT, nullable=False)  # 因子名称
    is_selected_by_default = Column(BOOLEAN, nullable=False)  # 是否默认被选中
    f_desc = Column(TEXT)  # 因子简介

class FundIndustryExposure(Base):
    '''基金行业暴露'''

    __tablename__ = 'fund_industry_exposure'

    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    ind_class_type = Column(Enum(IndClassType), nullable=False)  # 行业分类类型
    value = Column(TEXT, nullable=False)  # JSON格式的行业暴露数据


class FOFNav(FOFBase):
    '''fof净值数据'''

    __tablename__ = 'fof_nav'

    fof_id = Column(CHAR(16), primary_key=True)             # 产品ID
    manager_id = Column(CHAR(32), primary_key=True)         # 管理端ID
    datetime = Column(DATE, primary_key=True)               # 日期
    nav = Column(DOUBLE(asdecimal=False), nullable=False)   # 单位净值
    acc_net_value = Column(DOUBLE(asdecimal=False))         # 累计净值
    adjusted_nav = Column(DOUBLE(asdecimal=False))          # 复权净值
    ta_factor = Column(DOUBLE(asdecimal=False))             # 复权因子
    volume = Column(DOUBLE(asdecimal=False))                # 份额
    mv = Column(DOUBLE(asdecimal=False))                    # 市值
    ret = Column(DOUBLE(asdecimal=False))                   # 累计收益率


class FOFUnconfirmedNav(FOFBase):
    '''fof未确认净值数据'''

    __tablename__ = 'fof_nav_unconfirmed'

    fof_id = Column(CHAR(16), primary_key=True)             # 产品ID
    manager_id = Column(CHAR(32), primary_key=True)         # 管理端ID
    datetime = Column(DATE, primary_key=True)               # 日期
    nav = Column(DOUBLE(asdecimal=False), nullable=False)   # 单位净值
    acc_net_value = Column(DOUBLE(asdecimal=False))         # 累计净值
    v_nav = Column(DOUBLE(asdecimal=False))                 # 虚拟净值


class FOFNavPublic(FOFBase):
    '''fof净值数据发布表'''

    __tablename__ = 'fof_nav_public'

    fof_id = Column(CHAR(16), primary_key=True)             # 产品ID
    datetime = Column(DATE, primary_key=True)               # 日期
    nav = Column(DOUBLE(asdecimal=False), nullable=False)   # 单位净值
    acc_net_value = Column(DOUBLE(asdecimal=False))         # 累计净值
    adjusted_nav = Column(DOUBLE(asdecimal=False))          # 复权净值
    volume = Column(DOUBLE(asdecimal=False))                # 份额
    mv = Column(DOUBLE(asdecimal=False))                    # 市值
    ret = Column(DOUBLE(asdecimal=False))                   # 累计收益率


class FOFNavCalc(FOFBase):
    '''fof净值数据表(计算)'''

    __tablename__ = 'fof_nav_calc'

    fof_id = Column(CHAR(16), primary_key=True)             # 产品ID
    manager_id = Column(CHAR(32), primary_key=True)         # 管理端ID
    datetime = Column(DATE, primary_key=True)               # 日期
    nav = Column(DOUBLE(asdecimal=False), nullable=False)   # 单位净值
    acc_net_value = Column(DOUBLE(asdecimal=False))         # 累计净值
    adjusted_nav = Column(DOUBLE(asdecimal=False))          # 复权净值
    ta_factor = Column(DOUBLE(asdecimal=False))             # 复权因子
    volume = Column(DOUBLE(asdecimal=False))                # 总份额
    mv = Column(DOUBLE(asdecimal=False))                    # 总市值(净资产)
    interest = Column(DOUBLE(asdecimal=False))              # 当日利息
    management_fee = Column(DOUBLE(asdecimal=False))        # 当日管理费
    custodian_fee = Column(DOUBLE(asdecimal=False))         # 当日托管费
    administrative_fee = Column(DOUBLE(asdecimal=False))    # 当日行政服务费
    ret = Column(DOUBLE(asdecimal=False))                   # 累计收益率
    dy_per_share = Column(DOUBLE(asdecimal=False))          # 每股分红


class FOFPosition(FOFBase):
    '''fof持仓汇总数据'''

    __tablename__ = 'fof_position'

    fof_id = Column(CHAR(16), primary_key=True)  # 产品ID
    manager_id = Column(CHAR(32), primary_key=True)  # 管理端ID
    datetime = Column(DATE, primary_key=True)  # 日期
    position = Column(TEXT, nullable=False)  # 持仓信息
    total_current = Column(DOUBLE(asdecimal=False))  # 流动性资产及负债总计
    other_assets = Column(DOUBLE(asdecimal=False))  # 其他资产
    deposit_in_bank = Column(DOUBLE(asdecimal=False))  # 银行存款


class FOFInvestorData(FOFBase):
    '''fof投资人数据'''

    __tablename__ = 'fof_investor_data'

    fof_id = Column(CHAR(16), primary_key=True)  # 产品ID
    investor_id = Column(CHAR(16), primary_key=True)  # 客户合同号(投资人ID)
    manager_id = Column(CHAR(32))  # 管理端ID
    total_investment = Column(DOUBLE(asdecimal=False))  # 总投资额


class FOFPositionDetail(FOFBase):
    '''fof持仓详细数据'''

    __tablename__ = 'fof_position_detail'

    fof_id = Column(CHAR(16), primary_key=True)  # 产品ID
    manager_id = Column(CHAR(32), primary_key=True)  # 管理端ID
    datetime = Column(DATE, primary_key=True)  # 日期
    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    asset_type = Column(Enum(HoldingAssetType), nullable=False)  # 资产类型
    confirmed_nav = Column(TEXT, nullable=False)  # 确认日净值列表
    water_line = Column(TEXT, nullable=False)  # 水位线列表(累计净值)
    nav = Column(DOUBLE(asdecimal=False), nullable=False)  # 单位净值
    v_nav = Column(DOUBLE(asdecimal=False))  # 虚拟净值
    acc_nav = Column(DOUBLE(asdecimal=False), nullable=False)  # 累计净值
    total_shares = Column(DOUBLE(asdecimal=False), nullable=False)  # 总份额
    total_cost = Column(DOUBLE(asdecimal=False), nullable=False)  # 投资成本
    latest_mv = Column(DOUBLE(asdecimal=False), nullable=False)  # 当日市值
    total_ret = Column(DOUBLE(asdecimal=False), nullable=False)  # 累计收益
    total_rr = Column(DOUBLE(asdecimal=False), nullable=False)  # 累计收益率


class HedgeFundInvestorPurAndRedemp(FOFBase):
    '''私募客户申赎数据'''

    __tablename__ = 'hedge_fund_investor_pur_redemp'

    id = Column(Integer, primary_key=True, autoincrement=True)
    fof_id = Column(CHAR(16), primary_key=True)  # 产品ID
    manager_id = Column(CHAR(32), primary_key=True)  # 管理端ID
    datetime = Column(DATE, primary_key=True)  # 交易日期
    applied_date = Column(DATE)  # 申请日期
    deposited_date = Column(DATE)  # 交割日期
    event_type = Column(Integer, nullable=False)  # 交易类型
    investor_id = Column(CHAR(16), nullable=False)  # 客户合同号(投资人ID)
    net_asset_value = Column(DOUBLE(asdecimal=False), nullable=False)  # 单位净值
    acc_unit_value = Column(DOUBLE(asdecimal=False), nullable=False)  # 累计净值
    purchase_amount = Column(DOUBLE(asdecimal=False))  # 申认购金额(元)
    redemp_confirmed_amount = Column(DOUBLE(asdecimal=False))  # 赎回确认金额(元)
    raising_interest = Column(DOUBLE(asdecimal=False))  # 募集期利息(元)
    redemp_fee = Column(DOUBLE(asdecimal=False))  # 赎回费(元)
    carry_amount = Column(DOUBLE(asdecimal=False))  # 业绩报酬金额(元)
    share_changed = Column(DOUBLE(asdecimal=False), nullable=False)  # 份额变更
    share_after_trans = Column(DOUBLE(asdecimal=False), nullable=False)  # 交易后份额


class HedgeFundInvestorPurAndRedempSub(FOFBase):
    """私募客户申赎数据附表"""

    __tablename__ = 'hedge_fund_investor_pur_redemp_sub'

    id = Column(Integer, primary_key=True, autoincrement=True)
    main_id = Column(Integer)
    trade_confirm_url = Column(CHAR(127))                       # 交易确认单
    trade_apply_url = Column(CHAR(127))                         # 交易申请表
    remark = Column(CHAR(255))                                  # 备注


class HedgeFundInvestorDivAndCarry(FOFBase):
    '''私募客户分红/业绩计提数据'''

    __tablename__ = 'hedge_fund_investor_div_carry'

    id = Column(Integer, primary_key=True, autoincrement=True)
    fof_id = Column(CHAR(16), primary_key=True)  # 产品ID
    manager_id = Column(CHAR(32), primary_key=True)  # 管理端ID
    datetime = Column(DATE, primary_key=True)  # 交易日期
    applied_date = Column(DATE)  # 申请日期
    deposited_date = Column(DATE)  # 交割日期
    event_type = Column(Integer, nullable=False)  # 交易类型
    investor_id = Column(CHAR(16), nullable=False)  # 客户合同号(投资人ID)
    net_asset_value = Column(DOUBLE(asdecimal=False), nullable=False)  # 单位净值
    acc_unit_value = Column(DOUBLE(asdecimal=False), nullable=False)  # 累计净值
    total_dividend = Column(DOUBLE(asdecimal=False))  # 红利总额(元)
    cash_dividend = Column(DOUBLE(asdecimal=False))  # 现金分红金额(元)
    reinvest_amount = Column(DOUBLE(asdecimal=False))  # 再投资金额(元)
    carry_amount = Column(DOUBLE(asdecimal=False))  # 业绩报酬金额(元)
    share_changed = Column(DOUBLE(asdecimal=False), nullable=False)  # 份额变更
    share_after_trans = Column(DOUBLE(asdecimal=False), nullable=False)  # 交易后份额


class HedgeFundCustodianData(FOFBase):
    '''私募基金托管人数据'''

    __tablename__ = 'hedge_fund_custodian_data'

    id = Column(Integer, primary_key=True, autoincrement=True)
    fof_id = Column(CHAR(16), primary_key=True)  # 产品ID
    manager_id = Column(CHAR(32), primary_key=True)  # 管理端ID
    datetime = Column(DATE, primary_key=True)  # 交易日期
    nav = Column(DOUBLE(asdecimal=False), nullable=False)  # 单位净值
    equity = Column(DOUBLE(asdecimal=False), nullable=False)  # 资产净值(元)
    total_shares = Column(DOUBLE(asdecimal=False), nullable=False)  # 总份额
    other_assets = Column(DOUBLE(asdecimal=False), nullable=False)  # 其他资产
    deposit_in_bank = Column(DOUBLE(asdecimal=False))  # 银行存款
    interest_receivable = Column(DOUBLE(asdecimal=False))  # 应收利息
    pur_amt_receivable = Column(DOUBLE(asdecimal=False))  # 应收申购款
    sec_clearing_fund = Column(DOUBLE(asdecimal=False))  # 证券清算款
    dividends_receivable = Column(DOUBLE(asdecimal=False))  # 应收股利
    payable_to_managers = Column(DOUBLE(asdecimal=False))  # 应付管理人报酬
    payable_to_custodian = Column(DOUBLE(asdecimal=False))  # 应付托管费
    payable_to_admin_service = Column(DOUBLE(asdecimal=False))  # 应付行政服务费
    mutual_funds_data = Column(TEXT)  # 持仓公募信息
    hedge_funds_data = Column(TEXT)  # 持仓私募信息
    url = Column(CHAR(127))  # 



from .message_sender import MessageSender
import datetime
import requests
import json
import time
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
# from selenium.webdriver.support.ui import WebDriverWait
# from selenium.webdriver.support import expected_conditions as EC
# from selenium.webdriver.common.by import By

class ImageMessageSender(MessageSender):

    def __init__(self):
        super(ImageMessageSender, self).__init__()

        driver_path = '/tool/chromedriver'
        options = Options()
        options.add_argument('--headless')
        options.add_argument('--no-sandbox')
        options.add_argument('--hide-scrollbars')
        self.driver = webdriver.Chrome(executable_path=driver_path, options=options)
        width = 600
        height = 1150
        self.driver.set_window_size(width, height)
        self.driver.implicitly_wait(10)

    def capture_webpage(self, webpage_url):
        self.driver.get(webpage_url)
        time.sleep(1)
        # WebDriverWait(self.driver, 10).until(EC.presence_of_element_located((By.ID, 'tab-MONTH1')))
        filename = f'/tmp/{webpage_url.split("/")[-1]}.png'
        self.driver.save_screenshot(filename)
        return filename

    def send(self, msg_to, msg_content):
        # Send snapshot of webpage and url
        # doc: https://open.work.weixin.qq.com/api/doc/90000/90135/90236
        access_token = self.get_access_token()
        target_url = f'https://api.weixin.qq.com/cgi-bin/message/custom/send?access_token={access_token}'


        filename = self.capture_webpage(msg_content)
        files = {'file': ('snapshot.png', open(filename, 'rb'), 'image/png', {'Expires': '0'})}

        # upload image and get media_id
        # doc: https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/New_temporary_materials.html
        upload_url = f'https://api.weixin.qq.com/cgi-bin/media/upload?access_token={access_token}&type=image'
        res = requests.post(url=upload_url, files=files, timeout=30)
        data = json.loads(res.text)
        if 'errcode' in data:
            print(f'Failed to upload image, errmsg: {data["errmsg"]}')
            return
        media_id = data['media_id']

        # Send image
        message = {
            'touser': '|'.join(msg_to) if type(msg_to) is list else msg_to,
            'msgtype': 'image',
            'image' : {'media_id' : media_id}
        }
        res = requests.post(url=target_url, data=json.dumps(message), timeout=20)

        print(f'msg: {message} {datetime.datetime.now()}')



import pandas as pd
from .data_tables import StocksDataTables
from ..api.raw import RawDataApi, RawDatabaseConnector
from ..fund.derived.stocks.compute_stock import compute_stock


class StockDataManager:

    def __init__(self):
        self.sdt = StocksDataTables()
        self.stock_info = pd.DataFrame([])

    def collect(self, stock_id=None):
        self.stock_info = self.read_stock_info()
        if stock_id:
            params = [{'stock_id': stock_id}]
        else:
            params = [{'stock_id': stock_id} for stock_id in self.stock_info['stock_id']]

        for i in params:
            data = compute_stock(i)
            print(i['stock_id'], 'Done!')
            self.sdt.stocks.update(data)

    def pickle_save(self, file_path=None):
        print(self.sdt.stocks.keys())
        if not file_path:
            file_path = './stock_manager'
        import pickle
        with open(file_path, 'wb') as fp:
            pickle.dump(self, fp)

    def read_stock_info(self):
        df = RawDataApi().get_em_stock_info()
        if not isinstance(df, pd.DataFrame) and not df:
            print(df)
            raise Exception('No stock info!')
        return df




import json
import boto3
import decimal
import os
import concurrent.futures
import time

def decimal_form(number):
    return decimal.Decimal(number).quantize(decimal.Decimal("0.0000"))

def recursion_decimal_dict(dic):
    for key in dic:
        if isinstance(dic[key], dict):
            recursion_decimal_dict(dic[key])
        else:
            if isinstance(dic[key], list):
                if isinstance(dic[key][1], float):
                    dic[key] = [ decimal_form(_) for _ in dic[key] ]
                elif isinstance(dic[key][0], dict):
                    dic[key] = [ recursion_decimal_dict(_) for _ in dic[key] ]
            
            elif isinstance(dic[key], float):
                dic[key] = decimal_form(dic[key])
                
                
            else:
                pass
    return dic

def process_bk_result_item(json_name:str='bk_20100101_500.json', folder:str='./'):
    bk_id = json_name.split('.')[0].split('_')[2]
    with open(folder+'/'+json_name,'r') as f:
        result_dic = json.load(f)
    result_dic['ef_id'] = bk_id
    return result_dic

def create_ef_table(version):
    bk_table_name = 'effective_frontier_bk' + '_v' + str(version)
    dynamodb = boto3.resource('dynamodb')
    dynamodb.create_table(
        TableName=bk_table_name,
        KeySchema=[
            {
                'AttributeName': 'ef_id',
                'KeyType': 'HASH' 
            }
            ],
        AttributeDefinitions=[
            {
                'AttributeName': 'ef_id',
                'AttributeType': 'S'
            }
        ],
        BillingMode='PAY_PER_REQUEST',
    )

def save_bk_item(ef_bk_table, json_name, folder, json_list):
    bk_table = boto3.resource('dynamodb').Table(ef_bk_table)
    bk_result = process_bk_result_item(json_name, folder)
    bk_result = recursion_decimal_dict(bk_result)
    bk_table.put_item(Item=bk_result)
    if json_list.index(json_name) % 1000 == 0:
        print(json_name)

def loop_save(version, folder):
    ef_bk_table = 'effective_frontier_bk' + '_v' + str(version)
    json_list = os.listdir(folder)
    j = json_list
    h = round(len(j) / 100 * 8.7 / 60 / 60, 1)
    print(f'cost {h} hours ')
    time0 = time.time()
    with concurrent.futures.ProcessPoolExecutor(max_workers=3) as executor:
        future_to_bti = {executor.submit(save_bk_item, ef_bk_table, json_name, folder, json_list): json_name for json_name in j}
        for future in concurrent.futures.as_completed(future_to_bti):
            bt_i = future_to_bti[future]
            try:
                data = future.result()
            except Exception as exc:
                print(f'{bt_i} generated an exception: {exc}')
            else:
                pass   
    time1 = time.time()
    print(time1 - time0)

if __name__ == "__main__":

    # inputs
    version = 0
    folder = './frontier_result'
    #create_ef_table(version)
    #loop_save(version, folder)





import datetime
from typing import List, Tuple, Optional, Dict
import pandas as pd
import traceback
import numpy as np

from ...api.basic import BasicDataApi, BasicDatabaseConnector, FundNav
from ...api.derived import DerivedDataApi, DerivedDatabaseConnector
from ...view.derived_models import FundManagerInfo
from .derived_data_helper import DerivedDataHelper


class FundManagerProfitProcess:

    def __init__(self, data_helper: DerivedDataHelper):
        self._data_helper = data_helper

    def init(self, start_date: str, end_date: str):
        self._end_date = end_date
        self._end_date_dt = pd.to_datetime(end_date).date()
        self._fund_info: pd.DataFrame = DerivedDataApi().get_fund_manager_info()
        self._result = pd.DataFrame([])

    def calc_fund_profit(self, start_date, end_date, fund_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            start_nav = db_session.query(FundNav.adjusted_net_value).filter(
                FundNav.fund_id == fund_id,
                FundNav.datetime >= start_date,
            ).order_by(FundNav.datetime.asc()).first()
            end_nav = db_session.query(FundNav.adjusted_net_value).filter(
                FundNav.fund_id == fund_id,
                FundNav.datetime <= end_date,
            ).order_by(FundNav.datetime.desc()).first()
            if not start_nav or not end_nav:
                return None
            
            try:
                profit = end_nav[0] / start_nav[0] - 1
            except Exception as e:
                print(start_nav, end_nav)
                print(traceback.format_exc())
                return None
            return profit

    def calc(self):
        print(self._fund_info)
        with DerivedDatabaseConnector().managed_session() as db_session:
            for i in self._fund_info.index:
                profit = self.calc_fund_profit(
                    self._fund_info.loc[i, 'start_date'],
                    self._fund_info.loc[i, 'end_date'],
                    self._fund_info.loc[i, 'fund_id'],
                )
                db_session.query(FundManagerInfo).filter(
                    FundManagerInfo.fund_id == self._fund_info.loc[i, 'fund_id'],
                    FundManagerInfo.start_date == self._fund_info.loc[i, 'start_date'],
                    FundManagerInfo.mng_id == self._fund_info.loc[i, 'mng_id'],
                ).update(
                    {'profit': profit}
                )
                db_session.commit()
                print(i, self._fund_info.loc[i, 'fund_id'])
        print(self._result)

    def process(self, start_date: str, end_date: str, end_date_dt: datetime.date) -> List[str]:
        print(f'fund manager profit update on the last day of week {end_date_dt}')
        failed_tasks = []
        try:
            self.init(start_date, end_date)
            self.calc()
            # self._data_helper._upload_derived(self._result, FundManagerInfo.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('FundManagerProfitProcess')
        return failed_tasks


if __name__ == '__main__':
    date = '20200930'
    fsg = FundManagerProfitProcess(DerivedDataHelper())
    fsg.process(date, date, pd.to_datetime(date).date())
    # FundScoreGroupProcess.update_history('20201010', 5)



import pandas as pd
import numpy as np
import traceback
import json
import math
from ...api.raw import *
from ....util.calculator import *
from ...view.view_models import HFFundDailyCollection

class HFFundDailyCollectionProcessor(object):
    
    def __init__(self):
        self.raw_api = RawDataApi()
        self.fund_info = self.raw_api.get_hf_fund_info()
        self.fund_nav = self.raw_api.get_hf_fund_nav_dt()
        self.fund_nav = self.fund_nav.pivot_table(index='datetime',values='nav',columns='fund_id')
    
    def get_fund_price_start_and_end(self):
        result = []
        for fund_id in self.fund_nav:
            _df = self.fund_nav[[fund_id]].dropna()
            dic = {
                'fund_id':fund_id,
                'price_begin_date':_df.index[0],
                'price_end_date':_df.index[-1]
            }
            result.append(dic)
        df = pd.DataFrame(result)
        return df

    def get_fund_indicator(self):
        dic = {
            'annualized_ret':'annual_ret',
            'annualized_vol':'annual_vol',
            'sharpe':'sharpe_ratio',
            'fund_id':'fund_id',
        }
        result = []
        for fund_id in self.fund_nav:
            _df = self.fund_nav[fund_id].dropna()
            stats = Calculator.get_stat_result(dates=_df.index,values=_df.values,frequency='1D',risk_free_rate=0.025).__dict__
            stats['fund_id'] = fund_id
            result.append(stats)
        df = pd.DataFrame(result)
        df = df[dic.keys()].rename(columns=dic)
        return df

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')


    def _process(self):
        try:
            fund_info = self.fund_info.set_index('fund_id')
            fund_start_end_date_df = self.get_fund_price_start_and_end().set_index('fund_id')
            fund_indicator = self.get_fund_indicator().set_index('fund_id')

            df = pd.concat([fund_info, fund_start_end_date_df, fund_indicator], axis=1, sort=False)
            df = df.reset_index().rename(columns={'index':'fund_id'}).replace({np.inf: np.nan, -np.inf: np.nan})
            self.append_data(HFFundDailyCollection.__tablename__, df)
            return True
        
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def process(self):
        failed_tasks = []
        if not self._process():
            failed_tasks.append('hf_fund_daily_collection')
        return failed_tasks


if __name__ == '__main__':
    HFFundDailyCollectionProcessor().process()




from typing import Dict

import numpy as np
import pandas as pd

from .factor_types import LiquidityFactor, MixFactorBase
from .basic_factors import TurnoverFactor
from .utils import get_day


class TurnoverDateFactor(LiquidityFactor):
    # 换手率（X日）
    def __init__(self, num):
        super().__init__('t'+str(num)+'d')
        self.num = num
        self._deps.add(TurnoverFactor())

    def calc(self):
        tov = TurnoverFactor().get()
        ls = tov.index  # trading_day_list
        fac = tov.copy()
        for i in range(0, len(tov)):
            try:
                temp = get_day(self.num, ls, tov.index[i])
                temp2 = get_day(0, ls, tov.index[i])
                fac.iloc[i] = tov.loc[temp:temp2].mean()
            except Exception:
                fac.iloc[i] = np.nan
        self._factor = fac


class Turnover30DFactor(TurnoverDateFactor):
    # 换手率（30日）
    def __init__(self):
        super().__init__(30)


class Turnover60DFactor(TurnoverDateFactor):
    # 换手率（60日）
    def __init__(self):
        super().__init__(60)


class Turnover90DFactor(TurnoverDateFactor):
    # 换手率（90日）
    def __init__(self):
        super().__init__(90)


class MixLiquidityFactor(MixFactorBase, LiquidityFactor):
    # 合成因子
    def __init__(self):
        super(MixFactorBase, self).__init__('mixl')
        self._factor: Dict[pd.DataFrame] = {}
        self._deps.add(Turnover30DFactor())
        self._deps.add(Turnover60DFactor())
        self._deps.add(Turnover90DFactor())

    def calc(self, universe: str):
        self._factor[universe] = Turnover30DFactor().get_normalized(universe) + Turnover60DFactor().get_normalized(universe) + Turnover90DFactor().get_normalized(universe)
